#include <stdio.h>
#include <stdlib.h>
#include "alignment_simd.h"

/*
#define DEBUG
*/

#ifdef __SSE2__
#include <emmintrin.h>
#define EPU8_TYPE(ins)  ins##_epu8
#define EPI8_TYPE(ins)  ins##_epi8
#define EPI32_TYPE(ins) ins##_epi32
#define EPI64_TYPE(ins) ins##_epi64
#define SI128_TYPE(ins) ins##_si128
#define VTYPE128           __m128i

typedef union {
  VTYPE128 v;
  unsigned long long ull[sizeof(VTYPE128)/sizeof(unsigned long long)];
  unsigned int       ui[sizeof(VTYPE128)/sizeof(unsigned int)];
  unsigned char      uc[sizeof(VTYPE128)/sizeof(unsigned char)];
}  __attribute__ ((aligned (16))) vector128_t ;

#endif




#ifdef __SSE__
#include <xmmintrin.h>

#define PU8_TYPE(ins)  ins##_pu8
#define PI8_TYPE(ins)  ins##_pi8
#define PI32_TYPE(ins) ins##_pi32
#define SI64_TYPE(ins) ins##_si64
#define VTYPE64           __m64

typedef union {
  VTYPE64 v;
  unsigned long long ull[sizeof(VTYPE64)/sizeof(unsigned long long)];
  unsigned int       ui[sizeof(VTYPE64)/sizeof(unsigned int)];
  unsigned char      uc[sizeof(VTYPE64)/sizeof(unsigned char)];
}  __attribute__ ((aligned (16))) vector64_t ;

#endif








inline int alignment_sse2__compatible_proc() {
  unsigned long int _ax = 0, _bx = 0, _cx = 0, _dx = 0;
  unsigned long int _op = 1;
#if defined(__x86_64__)
  asm volatile (
                "pushq %%rbx" "\n\t"
                "cpuid" "\n\t"
                "movq %%rbx,%1" "\n\t"
                "popq %%rbx" "\n\t"
                : "=a" (_ax), "=m" (_bx), "=c" (_cx), "=d" (_dx)
                : "a" (_op)
                );
#else
  asm volatile (
                "pushl %%ebx" "\n\t"
                "cpuid" "\n\t"
                "movl %%ebx,%1" "\n\t"
                "popl %%ebx" "\n\t"
                : "=a" (_ax), "=m" (_bx), "=c" (_cx), "=d" (_dx)
                : "a" (_op)
                );
#endif
  printf("* compatible sse2 ? %s\n",_dx & 1<<26 ? "yes":"no");
  return (_dx & 1<<26) != 0;
}


inline int alignment_sse__compatible_proc() {
  unsigned long int _ax = 0, _bx = 0, _cx = 0, _dx = 0;
  unsigned long int _op = 1;
#if defined(__x86_64__)
  asm volatile (
                "pushq %%rbx" "\n\t"
                "cpuid" "\n\t"
                "movq %%rbx,%1" "\n\t"
                "popq %%rbx" "\n\t"
                : "=a" (_ax), "=m" (_bx), "=c" (_cx), "=d" (_dx)
                : "a" (_op)
                );
#else
  asm volatile (
                "pushl %%ebx" "\n\t"
                "cpuid" "\n\t"
                "movl %%ebx,%1" "\n\t"
                "popl %%ebx" "\n\t"
                : "=a" (_ax), "=m" (_bx), "=c" (_cx), "=d" (_dx)
                : "a" (_op)
                );
#endif
  printf("* compatible sse ? %s\n",_dx & 1<<25 ? "yes":"no");
  return (_dx & 1<<25) != 0;
}



/*
 * macros and global variables (no need to export because not usefull ouside this .c file)
 */

/* sse2 macros */

#ifdef __SSE2__

#define NEXTREADSEQ_QUAD128(inout_byte,inout_vector_nbletters,                              \
                            inout_vector,out_vtype_vLA) {                                   \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ui[0] = inout_vector.ui[1] = inout_vector.ui[2] = inout_vector.ui[3] =   \
        *((unsigned int *)(inout_byte));                                                    \
      inout_vector_nbletters = 16;                                                          \
      inout_byte += 4;                                                                      \
     }                                                                                      \
     out_vtype_vLA  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                   \
     inout_vector.v = EPI32_TYPE(_mm_srli)(inout_vector.v,2);                               \
     inout_vector_nbletters--;                                                              \
}

#define NEXTREADSEQ_PAIR128(inout_byte,inout_vector_nbletters,                              \
                            inout_vector,out_vtype_vLA) {                                   \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = inout_vector.ull[1] = *((unsigned long long *)(inout_byte));    \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
     }                                                                                      \
     out_vtype_vLA  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                   \
     inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                               \
     inout_vector_nbletters--;                                                              \
}

#define NEXTREADSEQ_SINGLE128(inout_byte,inout_vector_nbletters,                            \
                              inout_vector,out_vtype_vLA) {                                 \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte));                          \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
     }                                                                                      \
     out_vtype_vLA  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                   \
     inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                               \
     inout_vector_nbletters--;                                                              \
}


#define NEXTGENOSEQ_QUAD128(in_subpos_a,inout_byte_a,inout_vector_nbletters_a,              \
                            in_subpos_b,inout_byte_b,inout_vector_nbletters_b,              \
                            in_subpos_c,inout_byte_c,inout_vector_nbletters_c,              \
                            in_subpos_d,inout_byte_d,inout_vector_nbletters_d,              \
                            inout_vector,out_vtype_vLB) {                                   \
    if (!inout_vector_nbletters_a) {                                                        \
      inout_vector.ui[0] = *((unsigned long long *)(inout_byte_a));                         \
      inout_vector_nbletters_a = 16;                                                        \
      inout_byte_a += 4;                                                                    \
      if (in_subpos_a) {                                                                    \
        inout_vector.ui[0]      >>= in_subpos_a << 1;                                       \
        inout_vector_nbletters_a -= in_subpos_a;                                            \
        in_subpos_a               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    if (!inout_vector_nbletters_b) {                                                        \
      inout_vector.ui[1] = *((unsigned long long *)(inout_byte_b));                         \
      inout_vector_nbletters_b = 16;                                                        \
      inout_byte_b += 4;                                                                    \
      if (in_subpos_b) {                                                                    \
        inout_vector.ui[1]      >>= in_subpos_b << 1;                                       \
        inout_vector_nbletters_b -= in_subpos_b;                                            \
        in_subpos_b               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    if (!inout_vector_nbletters_c) {                                                        \
      inout_vector.ui[2] = *((unsigned long long *)(inout_byte_c));                         \
      inout_vector_nbletters_c = 16;                                                        \
      inout_byte_c += 4;                                                                    \
      if (in_subpos_c) {                                                                    \
        inout_vector.ui[2]      >>= in_subpos_c << 1;                                       \
        inout_vector_nbletters_c -= in_subpos_c;                                            \
        in_subpos_c               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    if (!inout_vector_nbletters_d) {                                                        \
      inout_vector.ui[3] = *((unsigned long long *)(inout_byte_d));                         \
      inout_vector_nbletters_d = 16;                                                        \
      inout_byte_d += 4;                                                                    \
      if (in_subpos_d) {                                                                    \
        inout_vector.ui[3]      >>= in_subpos_d << 1;                                       \
        inout_vector_nbletters_d -= in_subpos_d;                                            \
        in_subpos_d               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI32_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
    inout_vector_nbletters_c--;                                                             \
    inout_vector_nbletters_d--;                                                             \
}


#define NEXTGENOSEQ_PAIR128(in_subpos_a,inout_byte_a,inout_vector_nbletters_a,              \
                            in_subpos_b,inout_byte_b,inout_vector_nbletters_b,              \
                            inout_vector,out_vtype_vLB) {                                   \
    if (!inout_vector_nbletters_a) {                                                        \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte_a));                        \
      inout_vector_nbletters_a = 32;                                                        \
      inout_byte_a += 8;                                                                    \
      if (in_subpos_a) {                                                                    \
        inout_vector.ull[0]      >>= in_subpos_a << 1;                                      \
        inout_vector_nbletters_a -= in_subpos_a;                                            \
        in_subpos_a               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    if (!inout_vector_nbletters_b) {                                                        \
      inout_vector.ull[1] = *((unsigned long long *)(inout_byte_b));                        \
      inout_vector_nbletters_b = 32;                                                        \
      inout_byte_b += 8;                                                                    \
      if (in_subpos_b) {                                                                    \
        inout_vector.ull[1]      >>= in_subpos_b << 1;                                      \
        inout_vector_nbletters_b -= in_subpos_b;                                            \
        in_subpos_b               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
}

#define NEXTGENOSEQ_SINGLE128(in_subpos,inout_byte,inout_vector_nbletters,                  \
                              inout_vector,out_vtype_vLB) {                                 \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte));                          \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
      if (in_subpos) {                                                                      \
        inout_vector.ull[0]      >>= in_subpos << 1;                                        \
        inout_vector_nbletters -= in_subpos;                                                \
        in_subpos               = 0;                                                        \
      }                                                                                     \
    }                                                                                       \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters--;                                                               \
}


#define NEXTGENOSEQ_NOSUB_QUAD128(inout_byte_a,inout_vector_nbletters_a,                    \
                                  inout_byte_b,inout_vector_nbletters_b,                    \
                                  inout_byte_c,inout_vector_nbletters_c,                    \
                                  inout_byte_d,inout_vector_nbletters_d,                    \
                                  inout_vector,out_vtype_vLB) {                             \
    if (!inout_vector_nbletters_a) {                                                        \
      inout_vector.ui[0] = *((unsigned long long *)(inout_byte_a));                         \
      inout_vector_nbletters_a = 16;                                                        \
      inout_byte_a += 4;                                                                    \
    }                                                                                       \
    if (!inout_vector_nbletters_b) {                                                        \
      inout_vector.ui[1] = *((unsigned long long *)(inout_byte_b));                         \
      inout_vector_nbletters_b = 16;                                                        \
      inout_byte_b += 4;                                                                    \
    }                                                                                       \
    if (!inout_vector_nbletters_c) {                                                        \
      inout_vector.ui[2] = *((unsigned long long *)(inout_byte_c));                         \
      inout_vector_nbletters_c = 16;                                                        \
      inout_byte_c += 4;                                                                    \
    }                                                                                       \
    if (!inout_vector_nbletters_d) {                                                        \
      inout_vector.ui[3] = *((unsigned long long *)(inout_byte_d));                         \
      inout_vector_nbletters_d = 16;                                                        \
      inout_byte_d += 4;                                                                    \
    }                                                                                       \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI32_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
    inout_vector_nbletters_c--;                                                             \
    inout_vector_nbletters_d--;                                                             \
}

#define NEXTGENOSEQ_NOSUB_PAIR128(inout_byte_a,inout_vector_nbletters_a,                    \
                                  inout_byte_b,inout_vector_nbletters_b,                    \
                                  inout_vector,out_vtype_vLB) {                             \
    if (!inout_vector_nbletters_a) {                                                        \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte_a));                        \
      inout_vector_nbletters_a = 32;                                                        \
      inout_byte_a += 8;                                                                    \
    }                                                                                       \
    if (!inout_vector_nbletters_b) {                                                        \
      inout_vector.ull[1] = *((unsigned long long *)(inout_byte_b));                        \
      inout_vector_nbletters_b = 32;                                                        \
      inout_byte_b += 8;                                                                    \
    }                                                                                       \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
}

#define NEXTGENOSEQ_NOSUB_SINGLE128(inout_byte,inout_vector_nbletters,                      \
                                    inout_vector,out_vtype_vLB) {                           \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte));                          \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
    }                                                                                       \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters--;                                                               \
}



#define NEXTGENOSEQ_NOSUB_NOUP_QUAD128(inout_vector_nbletters_a,                            \
                                       inout_vector_nbletters_b,                            \
                                       inout_vector_nbletters_c,                            \
                                       inout_vector_nbletters_d,                            \
                                       inout_vector,out_vtype_vLB) {                        \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI32_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
    inout_vector_nbletters_c--;                                                             \
    inout_vector_nbletters_d--;                                                             \
}

#define NEXTGENOSEQ_NOSUB_NOUP_PAIR128(inout_vector_nbletters_a,inout_vector_nbletters_b,   \
                                       inout_vector,out_vtype_vLB) {                        \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
}

#define NEXTGENOSEQ_NOSUB_NOUP_SINGLE128(inout_vector_nbletters,                            \
                                         inout_vector,out_vtype_vLB) {                      \
    out_vtype_vLB  = SI128_TYPE(_mm_and)(inout_vector.v,vBufferMask128);                    \
    inout_vector.v = EPI64_TYPE(_mm_srli)(inout_vector.v,2);                                \
    inout_vector_nbletters--;                                                               \
}



VTYPE128  vThreshold128     __attribute__ ((aligned (16))),
          vMatchS128        __attribute__ ((aligned (16))),
          vMismatchS128     __attribute__ ((aligned (16))),
          vIndelOpenS128    __attribute__ ((aligned (16))),
          vIndelExtendsS128 __attribute__ ((aligned (16))),
          v0128             __attribute__ ((aligned (16))),
          vBufferMask128    __attribute__ ((aligned (16)));
VTYPE128  vMsk128[256]      __attribute__ ((aligned (16)));

#endif



/* sse macros */

#ifdef __SSE__

#define NEXTREADSEQ_PAIR64(inout_byte,inout_vector_nbletters,                               \
                           inout_vector,out_vtype_vLA) {                                    \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ui[0] = inout_vector.ui[1] = *((unsigned int *)(inout_byte));            \
      inout_vector_nbletters = 16;                                                          \
      inout_byte += 4;                                                                      \
    }                                                                                       \
    out_vtype_vLA  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = PI32_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters--;                                                               \
}

#define NEXTREADSEQ_SINGLE64(inout_byte,inout_vector_nbletters,                             \
                             inout_vector,out_vtype_vLA) {                                  \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte));                          \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
    }                                                                                       \
    out_vtype_vLA  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = SI64_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters--;                                                               \
}


#define NEXTGENOSEQ_PAIR64(in_subpos_a,inout_byte_a,inout_vector_nbletters_a,               \
                           in_subpos_b,inout_byte_b,inout_vector_nbletters_b,               \
                           inout_vector,out_vtype_vLB) {                                    \
    if (!inout_vector_nbletters_a) {                                                        \
      inout_vector.ui[0] = *((unsigned int *)(inout_byte_a));                               \
      inout_vector_nbletters_a = 16;                                                        \
      inout_byte_a += 4;                                                                    \
      if (in_subpos_a) {                                                                    \
        inout_vector.ui[0]     >>= in_subpos_a << 1;                                        \
        inout_vector_nbletters_a -= in_subpos_a;                                            \
        in_subpos_a               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    if (!inout_vector_nbletters_b) {                                                        \
      inout_vector.ui[1] = *((unsigned int *)(inout_byte_b));                               \
      inout_vector_nbletters_b = 16;                                                        \
      inout_byte_b += 4;                                                                    \
      if (in_subpos_b) {                                                                    \
        inout_vector.ui[1]      >>= in_subpos_b << 1;                                       \
        inout_vector_nbletters_b -= in_subpos_b;                                            \
        in_subpos_b               = 0;                                                      \
      }                                                                                     \
    }                                                                                       \
    out_vtype_vLB  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = PI32_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
}

#define NEXTGENOSEQ_SINGLE64(in_subpos,inout_byte,inout_vector_nbletters,                   \
                              inout_vector,out_vtype_vLB) {                                 \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte));                          \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
      if (in_subpos) {                                                                      \
        inout_vector.ull[0]   >>= in_subpos << 1;                                           \
        inout_vector_nbletters -= in_subpos;                                                \
        in_subpos               = 0;                                                        \
      }                                                                                     \
    }                                                                                       \
    out_vtype_vLB  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = SI64_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters--;                                                               \
}

#define NEXTGENOSEQ_NOSUB_PAIR64(inout_byte_a,inout_vector_nbletters_a,                     \
                                 inout_byte_b,inout_vector_nbletters_b,                     \
                                 inout_vector,out_vtype_vLB) {                              \
    if (!inout_vector_nbletters_a) {                                                        \
      inout_vector.ui[0] = *((unsigned int *)(inout_byte_a));                               \
      inout_vector_nbletters_a = 16;                                                        \
      inout_byte_a += 4;                                                                    \
    }                                                                                       \
    if (!inout_vector_nbletters_b) {                                                        \
      inout_vector.ui[1] = *((unsigned int *)(inout_byte_b));                               \
      inout_vector_nbletters_b = 16;                                                        \
      inout_byte_b += 4;                                                                    \
    }                                                                                       \
    out_vtype_vLB  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = PI32_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
}

#define NEXTGENOSEQ_NOSUB_SINGLE64(inout_byte,inout_vector_nbletters,                       \
                                   inout_vector,out_vtype_vLB) {                            \
    if (!inout_vector_nbletters) {                                                          \
      inout_vector.ull[0] = *((unsigned long long *)(inout_byte));                          \
      inout_vector_nbletters = 32;                                                          \
      inout_byte += 8;                                                                      \
    }                                                                                       \
    out_vtype_vLB  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = SI64_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters--;                                                               \
}



#define NEXTGENOSEQ_NOSUB_NOUP_PAIR64(inout_vector_nbletters_a,inout_vector_nbletters_b,    \
                                      inout_vector,out_vtype_vLB) {                         \
    out_vtype_vLB  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = PI32_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters_a--;                                                             \
    inout_vector_nbletters_b--;                                                             \
}

#define NEXTGENOSEQ_NOSUB_NOUP_SINGLE64(inout_vector_nbletters,                             \
                                        inout_vector,out_vtype_vLB) {                       \
    out_vtype_vLB  = SI64_TYPE(_mm_and)(inout_vector.v,vBufferMask64);                      \
    inout_vector.v = SI64_TYPE(_mm_srli)(inout_vector.v,2);                                 \
    inout_vector_nbletters--;                                                               \
}

VTYPE64  vThreshold64     __attribute__ ((aligned (16))),
         vMatchS64        __attribute__ ((aligned (16))),
         vMismatchS64     __attribute__ ((aligned (16))),
         vIndelOpenS64    __attribute__ ((aligned (16))),
         vIndelExtendsS64 __attribute__ ((aligned (16))),
         v064             __attribute__ ((aligned (16))),
         vBufferMask64    __attribute__ ((aligned (16)));
VTYPE64  vMsk64[256]      __attribute__ ((aligned (16)));

#endif


unsigned int   prlength;



#ifdef __SSE2__

/**
 * SSE2 alignment init read function : modify the read length when needed (but must not be changed too frequently).
 * @param readlength gives the read length (number of nucleotides inside the read)
 */

inline void alignment_sse2__setlength_single(unsigned int readlength) {

  /* masking table to keep only good diagonals */
  prlength = (readlength+15) * 2;
  vMsk128[0] = EPI8_TYPE(_mm_set)(0xff,0x00,0x00,0x00,
                                  0x00,0x00,0x00,0x00,
                                  0x00,0x00,0x00,0x00,
                                  0x00,0x00,0x00,0x00);
#ifdef DEBUG
  {
    vector128_t Msk;
    Msk.v = vMsk128[0];
    fprintf(stdout,"[0]\t Msk:%.16llx%.16llx\n",(Msk.ull[1]),(Msk.ull[0]));
  }
#endif
  unsigned int l;
  for ( l = 1; l < prlength; l++ ) {
    /* middle mask */
    vMsk128[l] = vMsk128[l - 1];
    if (!(l & 1)) {
      /* mask at the end */
      if (l >= prlength - 15*2) {
        vMsk128[l] = SI128_TYPE(_mm_srli)(vMsk128[l],(1));
      } else {
        /* mask at the beginning */
        if ( l <= 15*2) {
          vMsk128[l] = SI128_TYPE(_mm_or)(vMsk128[l-1],SI128_TYPE(_mm_srli)(vMsk128[l],(1)));
        }
      }
    }
#ifdef DEBUG
    {
      vector128_t Msk;
      Msk.v = vMsk128[l];
      fprintf(stdout,"[0]\t Msk:%.16llx%.16llx\n",(Msk.ull[1]),(Msk.ull[0]));
    }
#endif
  }/* for l */
}



inline void alignment_sse2__setlength_pair(unsigned int readlength) {

  /* masking table to keep only good diagonals */
  prlength = (readlength+7) * 2;
  vMsk128[0] = EPI8_TYPE(_mm_set)(0xff,0x00,0x00,0x00,
                                  0x00,0x00,0x00,0x00,
                                  0xff,0x00,0x00,0x00,
                                  0x00,0x00,0x00,0x00);
#ifdef DEBUG
  {
    vector128_t Msk;
    Msk.v = vMsk128[0];
    fprintf(stdout,"[0]\t Msk:%.16llx,%.16llx\n",(Msk.ull[1]),(Msk.ull[0]));
  }
#endif
  unsigned int l;
  for ( l = 1; l < prlength; l++ ) {
    /* middle mask */
    vMsk128[l] = vMsk128[l - 1];
    if (!(l & 1)) {
      /* mask at the end */
      if (l >= prlength - 7*2) {
        vMsk128[l] = EPI64_TYPE(_mm_srli)(vMsk128[l],(1)*8);
      } else {
        /* mask at the beginning */
        if ( l <= 7*2) {
          vMsk128[l] = SI128_TYPE(_mm_or)(vMsk128[l-1],EPI64_TYPE(_mm_srli)(vMsk128[l],(1)*8));
        }
      }
    }
#ifdef DEBUG
    {
      vector128_t Msk;
      Msk.v = vMsk128[l];
      fprintf(stdout,"[0]\t Msk:%.16llx,%.16llx\n",(Msk.ull[1]),(Msk.ull[0]));
    }
#endif
  }/* for l */
}



inline void alignment_sse2__setlength_quad(unsigned int readlength) {

  /* masking table to keep only good diagonals */
  prlength = (readlength+3) * 2;
  vMsk128[0] = EPI8_TYPE(_mm_set)(0xff,0x00,0x00,0x00,
                                  0xff,0x00,0x00,0x00,
                                  0xff,0x00,0x00,0x00,
                                  0xff,0x00,0x00,0x00);
#ifdef DEBUG
  {
    vector128_t Msk;
    Msk.v = vMsk128[0];
    fprintf(stdout,"[0]\t Msk:%.8x,%.8x,%.8x,%.8x\n",(Msk.ui[3]),(Msk.ui[2]),(Msk.ui[1]),(Msk.ui[0]));
  }
#endif
  unsigned int l;
  for ( l = 1; l < prlength; l++ ) {
    /* middle mask */
    vMsk128[l] = vMsk128[l - 1];
    if (!(l & 1)) {
      /* mask at the end */
      if (l >= prlength - 3*2) {
        vMsk128[l] = EPI32_TYPE(_mm_srli)(vMsk128[l],(1)*8);
      } else {
        /* mask at the beginning */
        if ( l <= 3*2) {
          vMsk128[l] = SI128_TYPE(_mm_or)(vMsk128[l-1],EPI32_TYPE(_mm_srli)(vMsk128[l],(1)*8));
        }
      }
    }
#ifdef DEBUG
    {
      vector128_t Msk;
      Msk.v = vMsk128[l];
      fprintf(stdout,"[0]\t Msk:%.8x,%.8x,%.8x,%.8x\n",(Msk.ui[3]),(Msk.ui[2]),(Msk.ui[1]),(Msk.ui[0]));
    }
#endif
  }/* for l */
}








/**
 * SSE2 alignment init function : fix the scoring system and the length of the reads (must be called once before aligning)
 * @param match      inits the match score vector
 * @param mismatch   inits the mismatch penalty vector (positive value only)
 * @param gapopen    inits the gap penalty vector (positive value only)
 * @param gapextends inits the gap penalty vector (positive value only)
 * @param threshold  inits the scoring threshold (positive value only)
 * @param length     fixes the length of the reads that will be treated : this value can be changed @see alignment_sse2__setlength_pair
 *        (but must not be changed too frequently).
 */


inline void alignment_sse2__init_single(unsigned int match, unsigned int mismatch, unsigned int gapopen, unsigned int gapextends, unsigned int threshold, unsigned int readlength) {

  alignment_sse2__compatible_proc();

  /* init some vectors */
  v0128             =  EPI32_TYPE(_mm_set)(0, 0, 0, 0);
  vBufferMask128    =  EPI32_TYPE(_mm_set)(0, 0, 0, 3);
  vThreshold128     =  EPI8_TYPE(_mm_set)(threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold);
  vIndelOpenS128    =  EPI8_TYPE(_mm_set)(gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen);
  vIndelExtendsS128 =  EPI8_TYPE(_mm_set)(gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends);
  vMatchS128        =  EPI8_TYPE(_mm_set)(match, match, match, match,
                                          match, match, match, match,
                                          match, match, match, match,
                                          match, match, match, match);
  vMismatchS128     =  EPI8_TYPE(_mm_set)(mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch);

  alignment_sse2__setlength_single(readlength);
}


inline void alignment_sse2__init_pair(unsigned int match, unsigned int mismatch, unsigned int gapopen, unsigned int gapextends, unsigned int threshold, unsigned int readlength) {

  alignment_sse2__compatible_proc();

  /* init some vectors */
  v0128             =  EPI32_TYPE(_mm_set)(0, 0, 0, 0);
  vBufferMask128    =  EPI32_TYPE(_mm_set)(0, 3, 0, 3);
  vThreshold128     =  EPI8_TYPE(_mm_set)(threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold);
  vIndelOpenS128    =  EPI8_TYPE(_mm_set)(gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen);
  vIndelExtendsS128 =  EPI8_TYPE(_mm_set)(gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends);
  vMatchS128        =  EPI8_TYPE(_mm_set)(match, match, match, match,
                                          match, match, match, match,
                                          match, match, match, match,
                                          match, match, match, match);
  vMismatchS128     =  EPI8_TYPE(_mm_set)(mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch);

  alignment_sse2__setlength_pair(readlength);
}


inline void alignment_sse2__init_quad(unsigned int match, unsigned int mismatch, unsigned int gapopen, unsigned int gapextends, unsigned int threshold, unsigned int readlength) {

  alignment_sse2__compatible_proc();

  /* init some vectors */
  v0128             =  EPI32_TYPE(_mm_set)(0, 0, 0, 0);
  vBufferMask128    =  EPI32_TYPE(_mm_set)(3, 3, 3, 3);
  vThreshold128     =  EPI8_TYPE(_mm_set)(threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold,
                                          threshold, threshold, threshold, threshold);
  vIndelOpenS128    =  EPI8_TYPE(_mm_set)(gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen,
                                          gapopen, gapopen, gapopen, gapopen);
  vIndelExtendsS128 =  EPI8_TYPE(_mm_set)(gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends,
                                          gapextends, gapextends, gapextends, gapextends);
  vMatchS128        =  EPI8_TYPE(_mm_set)(match, match, match, match,
                                          match, match, match, match,
                                          match, match, match, match,
                                          match, match, match, match);
  vMismatchS128     =  EPI8_TYPE(_mm_set)(mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch,
                                          mismatch, mismatch, mismatch, mismatch);

  alignment_sse2__setlength_quad(readlength);
}




/**
 * SSE2 alignment align function : does a banded smith-waterman of the given read against two parts of the genome;
 * allows at most 3/4 (4/3) indels on each side.
 * @param genome is the compressed genome (first nucleotide is the lower bit of the first byte)
 * @param pos_genome_a gives the 'A' position (in term of nucleotides) to be aligned with the read (you must substract 3 or 4 potetial indels according to the hit position)
 * @param pos_genome_a gives the 'B' position (in term of nucleotides) to be aligned with the read (you must substract 3 or 4 potetial indels according to the hit position)
 * @param read is the compressed read (first nucleotide is the lower bit of the first byte)
 * @return 0 if none is aligned to reach the given threshold (what shoul happend most of the time),
 *         or a bitmask 1<<(A?0) | 1<<(B?1) if A or B align correctly.
 */


inline int alignment_sse2__align_single(unsigned char * genome,
                                        unsigned int * pos_genome,
                                        unsigned char * read) {
  VTYPE128 vA;
  VTYPE128 vB;

  VTYPE128 vMMax;
  VTYPE128 vM_old;
  VTYPE128 vM_old_old;
  VTYPE128 vI_old;

  unsigned int pos_genome_a = pos_genome[0];

  vector128_t     vector_genome_buffer __attribute__ ((aligned (16)));
  unsigned char *             byte_pos_genome = genome + (pos_genome_a >> 2);
  unsigned int                 sub_pos_genome = (pos_genome_a & 3);
  unsigned int vector_genome_buffer_nbletters = 0;

  vector128_t     vector_read_buffer   __attribute__ ((aligned (16)));
  unsigned int vector_read_buffer_nbletters = 0;


  vMMax      = SI128_TYPE(_mm_setzero)();
  vM_old     = SI128_TYPE(_mm_setzero)();
  vM_old_old = SI128_TYPE(_mm_setzero)();
  vI_old     = SI128_TYPE(_mm_setzero)();

  /* init vA : read sequence diagonal */
  {
    vA       = SI128_TYPE(_mm_setzero)();
  }

  /* init vB : genome sequence diagonal */
  {
    int d;
    NEXTGENOSEQ_SINGLE128(sub_pos_genome,byte_pos_genome,vector_genome_buffer_nbletters,
                          vector_genome_buffer,vB);
    for (d=1; d < 16; d++) {
      VTYPE128 vLB;
      NEXTGENOSEQ_NOSUB_NOUP_SINGLE128(vector_genome_buffer_nbletters,
                                       vector_genome_buffer,vLB);
      vB        = SI128_TYPE(_mm_slli)(vB,(1));
      vB        = SI128_TYPE(_mm_or)(vB,vLB);
    }
  }


  /* main loop */
  {
    unsigned int l;
    for (l=0; l<prlength; l++) {

      /* a) get the letters to be compared on the diagonal */
      if (l & 1) {
        VTYPE128 vLB;
        NEXTGENOSEQ_NOSUB_SINGLE128(byte_pos_genome,vector_genome_buffer_nbletters,
                                    vector_genome_buffer,vLB);
        vB  = SI128_TYPE(_mm_slli)(vB,(1));
        vB  = SI128_TYPE(_mm_or)(vB,vLB);
      } else {
        VTYPE128 vLA;
        NEXTREADSEQ_SINGLE128(read,vector_read_buffer_nbletters,
                              vector_read_buffer,vLA);
        vLA = SI128_TYPE(_mm_slli)(vLA,(15));
        vA  = SI128_TYPE(_mm_srli)(vA,(1));
        vA  = SI128_TYPE(_mm_or)(vA,vLA);
      }

#ifdef DEBUG
      {
        vector128_t A,B;
        A.v = vA; B.v = vB;
        fprintf(stdout,"[1]\t   A:%.16llx%.16llx\n",(A.ull[1]),(A.ull[0]));
        fprintf(stdout,"[1]\t   B:%.16llx%.16llx\n",(B.ull[1]),(B.ull[0]));
      }
#endif


      VTYPE128 vM;
      /* b) compute the matching score */
      {
        VTYPE128 vM_ab_MatchMask = EPI8_TYPE(_mm_cmpeq)(vA,vB);
        VTYPE128 vM_add = SI128_TYPE(_mm_and)(vM_ab_MatchMask,vMatchS128);
        VTYPE128 vM_sub = SI128_TYPE(_mm_andnot)(vM_ab_MatchMask,vMismatchS128);

#ifdef DEBUG
        {
          vector128_t S_a,S_s;
          S_a.v = vM_add; S_s.v = vM_sub;
          fprintf(stdout,"[1]\t S_a:%.16llx%.16llx\n",(S_a.ull[1]),(S_a.ull[0]));
          fprintf(stdout,"[1]\t S_s:%.16llx%.16llx\n",(S_s.ull[1]),(S_s.ull[0]));
        }
#endif


#ifdef DEBUG
        {
          vector128_t M_old_old;
          M_old_old.v = vM_old_old;
          fprintf(stdout,"[1]\t ooM:%.16llx%.16llx\n",(M_old_old.ull[1]),(M_old_old.ull[0]));
        }
#endif

        /* same diagonal with M_old_old */
        vM = EPU8_TYPE(_mm_adds)(vM_old_old,vM_add);
        vM = EPU8_TYPE(_mm_subs)(vM,vM_sub);
      }

#ifdef DEBUG
      {
        vector128_t M,M_old,I_old;
        M.v     = vM;
        M_old.v = vM_old;
        I_old.v = vI_old;
        fprintf(stdout,"[1]\t>  M:%.16llx%.16llx\n",(M.ull[1]),(M.ull[0]));
        fprintf(stdout,"[1]\t  oM:%.16llx%.16llx\n",(M_old.ull[1]),(M_old.ull[0]));
        fprintf(stdout,"[1]\t  oI:%.16llx%.16llx\n",(I_old.ull[1]),(I_old.ull[0]));
      }
#endif

      VTYPE128 vI;
      {
        /* shift */
        VTYPE128 vM_old_shifted;
        VTYPE128 vI_old_shifted;
        if (l & 1) {
          vM_old_shifted  = SI128_TYPE(_mm_slli)(vM_old,(1));
          vI_old_shifted  = SI128_TYPE(_mm_slli)(vI_old,(1));
        } else {
          vM_old_shifted  = SI128_TYPE(_mm_srli)(vM_old,(1));
          vI_old_shifted  = SI128_TYPE(_mm_srli)(vI_old,(1));
        }
        VTYPE128 vI_old_merge = EPU8_TYPE(_mm_max)(vI_old,vI_old_shifted);
        VTYPE128 vM_old_merge = EPU8_TYPE(_mm_max)(vM_old,vM_old_shifted);
        vI                    = EPU8_TYPE(_mm_subs)(vI_old_merge,vIndelExtendsS128);
        VTYPE128 vIstart      = EPU8_TYPE(_mm_subs)(vM_old_merge,vIndelOpenS128);
        vI                    = EPU8_TYPE(_mm_max)(vI,vIstart);
        vM                    = EPU8_TYPE(_mm_max)(vM,vI);
      }

#ifdef DEBUG
      {
        vector128_t M,I;
        M.v = vM;
        I.v = vI;
        fprintf(stdout,"[1]\t>  M:%.16llx%.16llx\n",(M.ull[1]),(M.ull[0]));
        fprintf(stdout,"[1]\t>  I:%.16llx%.16llx\n",(I.ull[1]),(I.ull[0]));
      }
#endif

      vM    = SI128_TYPE(_mm_and)(vM,vMsk128[l]);
      vMMax = EPU8_TYPE(_mm_max)(vMMax,vM);

#ifdef DEBUG
      {
        vector128_t Msk,M,Max;
        Msk.v = vMsk128[l];
        M.v   = vM;
        Max.v = vMMax;
        fprintf(stdout,"[1]\t>Msk:%.16llx%.16llx\n",(Msk.ull[1]),(Msk.ull[0]));
        fprintf(stdout,"[1]\t>  M:%.16llx%.16llx\n",(M.ull[1]),(M.ull[0]));
        fprintf(stdout,"[1]\t>Max:%.16llx%.16llx\n",(Max.ull[1]),(Max.ull[0]));
      }
#endif

      vM_old_old = vM_old;
      vM_old     = vM;
      vI_old     = vI;

#ifdef DEBUG
      fprintf(stdout,"[1]\t     --------------------------------\n");
#endif
    } /* l */
  }
  {
    VTYPE128 vThresholdMask = EPU8_TYPE(_mm_subs)(vMMax,vThreshold128);
    int result = 0;
    int x;
    for(x=0;x<(int)sizeof(VTYPE128)/(int)sizeof(unsigned long long int);x++) {
      unsigned long long int u =  ((vector128_t)vThresholdMask).ull[x];
        if (u != (unsigned long long int) 0) {
          result |= (1)<<x;
        }
    }
#ifdef DEBUG
        fprintf(stdout,"[1]\t     ================================\n");
#endif
    return result;
  }
}


inline int alignment_sse2__align_pair(unsigned char * genome,
                                      unsigned int * pos_genome,
                                      unsigned char * read) {
  VTYPE128 vA;
  VTYPE128 vB;

  VTYPE128 vMMax;
  VTYPE128 vM_old;
  VTYPE128 vM_old_old;
  VTYPE128 vI_old;

  unsigned int pos_genome_a = pos_genome[0];
  unsigned int pos_genome_b = pos_genome[1];

  vector128_t     vector_genome_buffer __attribute__ ((aligned (16)));

  unsigned char *             byte_pos_genome_a = genome + (pos_genome_a >> 2);
  unsigned int                 sub_pos_genome_a = (pos_genome_a & 3);
  unsigned int vector_genome_buffer_nbletters_a = 0;

  unsigned char *             byte_pos_genome_b = genome + (pos_genome_b >> 2);
  unsigned int                 sub_pos_genome_b = (pos_genome_b & 3);
  unsigned int vector_genome_buffer_nbletters_b = 0;


  vector128_t     vector_read_buffer     __attribute__ ((aligned (16)));

  unsigned int vector_read_buffer_nbletters = 0;


  vMMax      = SI128_TYPE(_mm_setzero)();
  vM_old     = SI128_TYPE(_mm_setzero)();
  vM_old_old = SI128_TYPE(_mm_setzero)();
  vI_old     = SI128_TYPE(_mm_setzero)();

  /* init vA : read sequence diagonal */
  {
    vA       = SI128_TYPE(_mm_setzero)();
  }

  /* init vB : genome sequence diagonal */
  {
    int d;
    NEXTGENOSEQ_PAIR128(sub_pos_genome_a,byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                        sub_pos_genome_b,byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                        vector_genome_buffer,vB);
    for (d=1; d < 8; d++) {
      VTYPE128 vLB;
      NEXTGENOSEQ_NOSUB_NOUP_PAIR128(vector_genome_buffer_nbletters_a,
                                     vector_genome_buffer_nbletters_b,
                                     vector_genome_buffer,vLB);
      vB        = EPI64_TYPE(_mm_slli)(vB,(1)*8);
      vB        = SI128_TYPE(_mm_or)(vB,vLB);
    }
  }


  /* main loop */
  {
    unsigned int l;
    for (l=0; l<prlength; l++) {

      /* a) get the letters to be compared on the diagonal */
      if (l & 1) {
        VTYPE128 vLB;
        NEXTGENOSEQ_NOSUB_PAIR128(byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                                  byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                                  vector_genome_buffer,vLB);
        vB  = EPI64_TYPE(_mm_slli)(vB,(1)*8);
        vB  = SI128_TYPE(_mm_or)(vB,vLB);
      } else {
        VTYPE128 vLA;
        NEXTREADSEQ_PAIR128(read,vector_read_buffer_nbletters,
                            vector_read_buffer,vLA);
        vLA = EPI64_TYPE(_mm_slli)(vLA,(7)*8);
        vA  = EPI64_TYPE(_mm_srli)(vA,(1)*8);
        vA  = SI128_TYPE(_mm_or)(vA,vLA);
      }

#ifdef DEBUG
      {
        vector128_t A,B;
        A.v = vA; B.v = vB;
        fprintf(stdout,"[1]\t   A:%.16llx,%.16llx\n",(A.ull[1]),(A.ull[0]));
        fprintf(stdout,"[1]\t   B:%.16llx,%.16llx\n",(B.ull[1]),(B.ull[0]));
      }
#endif


      VTYPE128 vM;
      /* b) compute the matching score */
      {
        VTYPE128 vM_ab_MatchMask = EPI8_TYPE(_mm_cmpeq)(vA,vB);
        VTYPE128 vM_add = SI128_TYPE(_mm_and)(vM_ab_MatchMask,vMatchS128);
        VTYPE128 vM_sub = SI128_TYPE(_mm_andnot)(vM_ab_MatchMask,vMismatchS128);

#ifdef DEBUG
        {
          vector128_t S_a,S_s;
          S_a.v = vM_add; S_s.v = vM_sub;
          fprintf(stdout,"[1]\t S_a:%.16llx,%.16llx\n",(S_a.ull[1]),(S_a.ull[0]));
          fprintf(stdout,"[1]\t S_s:%.16llx,%.16llx\n",(S_s.ull[1]),(S_s.ull[0]));
        }
#endif


#ifdef DEBUG
        {
          vector128_t M_old_old;
          M_old_old.v = vM_old_old;
          fprintf(stdout,"[1]\t ooM:%.16llx,%.16llx\n",(M_old_old.ull[1]),(M_old_old.ull[0]));
        }
#endif

        /* same diagonal with M_old_old */
        vM = EPU8_TYPE(_mm_adds)(vM_old_old,vM_add);
        vM = EPU8_TYPE(_mm_subs)(vM,vM_sub);
      }

#ifdef DEBUG
      {
        vector128_t M,M_old,I_old;
        M.v     = vM;
        M_old.v = vM_old;
        I_old.v = vI_old;
        fprintf(stdout,"[1]\t>  M:%.16llx,%.16llx\n",(M.ull[1]),(M.ull[0]));
        fprintf(stdout,"[1]\t  oM:%.16llx,%.16llx\n",(M_old.ull[1]),(M_old.ull[0]));
        fprintf(stdout,"[1]\t  oI:%.16llx,%.16llx\n",(I_old.ull[1]),(I_old.ull[0]));
      }
#endif

      VTYPE128 vI;
      {
        /* shift */
        VTYPE128 vM_old_shifted;
        VTYPE128 vI_old_shifted;
        if (l & 1) {
          vM_old_shifted  = EPI64_TYPE(_mm_slli)(vM_old,(1)*8);
          vI_old_shifted  = EPI64_TYPE(_mm_slli)(vI_old,(1)*8);
        } else {
          vM_old_shifted  = EPI64_TYPE(_mm_srli)(vM_old,(1)*8);
          vI_old_shifted  = EPI64_TYPE(_mm_srli)(vI_old,(1)*8);
        }
        VTYPE128 vI_old_merge = EPU8_TYPE(_mm_max)(vI_old,vI_old_shifted);
        VTYPE128 vM_old_merge = EPU8_TYPE(_mm_max)(vM_old,vM_old_shifted);
        vI                    = EPU8_TYPE(_mm_subs)(vI_old_merge,vIndelExtendsS128);
        VTYPE128 vIstart      = EPU8_TYPE(_mm_subs)(vM_old_merge,vIndelOpenS128);
        vI                    = EPU8_TYPE(_mm_max)(vI,vIstart);
        vM                    = EPU8_TYPE(_mm_max)(vM,vI);
      }

#ifdef DEBUG
      {
        vector128_t M,I;
        M.v = vM;
        I.v = vI;
        fprintf(stdout,"[1]\t>  M:%.16llx,%.16llx\n",(M.ull[1]),(M.ull[0]));
        fprintf(stdout,"[1]\t>  I:%.16llx,%.16llx\n",(I.ull[1]),(I.ull[0]));
      }
#endif

      vM    = SI128_TYPE(_mm_and)(vM,vMsk128[l]);
      vMMax = EPU8_TYPE(_mm_max)(vMMax,vM);

#ifdef DEBUG
      {
        vector128_t Msk,M,Max;
        Msk.v = vMsk128[l];
        M.v   = vM;
        Max.v = vMMax;
        fprintf(stdout,"[1]\t>Msk:%.16llx,%.16llx\n",(Msk.ull[1]),(Msk.ull[0]));
        fprintf(stdout,"[1]\t>  M:%.16llx,%.16llx\n",(M.ull[1]),(M.ull[0]));
        fprintf(stdout,"[1]\t>Max:%.16llx,%.16llx\n",(Max.ull[1]),(Max.ull[0]));
      }
#endif

      vM_old_old = vM_old;
      vM_old     = vM;
      vI_old     = vI;

#ifdef DEBUG
      fprintf(stdout,"[1]\t     ----------------,----------------\n");
#endif
    } /* l */
  }
  {
    VTYPE128 vThresholdMask = EPU8_TYPE(_mm_subs)(vMMax,vThreshold128);
    int result = 0;
    int x;
    for(x=0;x<(int)sizeof(VTYPE128)/(int)sizeof(unsigned long long int);x++) {
      unsigned long long int u =  ((vector128_t)vThresholdMask).ull[x];
        if (u != (unsigned long long int) 0) {
          result |= (1)<<x;
        }
    }
#ifdef DEBUG
        fprintf(stdout,"[1]\t     ================,================\n");
#endif
    return result;
  }
}


inline int alignment_sse2__align_quad(unsigned char * genome,
                                      unsigned int * pos_genome,
                                      unsigned char * read) {
  VTYPE128 vA;
  VTYPE128 vB;

  VTYPE128 vMMax;
  VTYPE128 vM_old;
  VTYPE128 vM_old_old;
  VTYPE128 vI_old;

  unsigned int pos_genome_a = pos_genome[0];
  unsigned int pos_genome_b = pos_genome[1];
  unsigned int pos_genome_c = pos_genome[2];
  unsigned int pos_genome_d = pos_genome[3];

  vector128_t     vector_genome_buffer __attribute__ ((aligned (16)));

  unsigned char *             byte_pos_genome_a = genome + (pos_genome_a >> 2);
  unsigned int                 sub_pos_genome_a = (pos_genome_a & 3);
  unsigned int vector_genome_buffer_nbletters_a = 0;

  unsigned char *             byte_pos_genome_b = genome + (pos_genome_b >> 2);
  unsigned int                 sub_pos_genome_b = (pos_genome_b & 3);
  unsigned int vector_genome_buffer_nbletters_b = 0;

  unsigned char *             byte_pos_genome_c = genome + (pos_genome_c >> 2);
  unsigned int                 sub_pos_genome_c = (pos_genome_c & 3);
  unsigned int vector_genome_buffer_nbletters_c = 0;

  unsigned char *             byte_pos_genome_d = genome + (pos_genome_d >> 2);
  unsigned int                 sub_pos_genome_d = (pos_genome_d & 3);
  unsigned int vector_genome_buffer_nbletters_d = 0;


  vector128_t     vector_read_buffer     __attribute__ ((aligned (16)));

  unsigned int vector_read_buffer_nbletters = 0;


  vMMax      = SI128_TYPE(_mm_setzero)();
  vM_old     = SI128_TYPE(_mm_setzero)();
  vM_old_old = SI128_TYPE(_mm_setzero)();
  vI_old     = SI128_TYPE(_mm_setzero)();

  /* init vA : read sequence diagonal */
  {
    vA       = SI128_TYPE(_mm_setzero)();
  }

  /* init vB : genome sequence diagonal */
  {
    int d;
    NEXTGENOSEQ_QUAD128(sub_pos_genome_a,byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                        sub_pos_genome_b,byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                        sub_pos_genome_c,byte_pos_genome_c,vector_genome_buffer_nbletters_c,
                        sub_pos_genome_d,byte_pos_genome_d,vector_genome_buffer_nbletters_d,
                        vector_genome_buffer,vB);
    for (d=1; d < 4; d++) {
      VTYPE128 vLB;
      NEXTGENOSEQ_NOSUB_NOUP_QUAD128(vector_genome_buffer_nbletters_a,
                                     vector_genome_buffer_nbletters_b,
                                     vector_genome_buffer_nbletters_c,
                                     vector_genome_buffer_nbletters_d,
                                     vector_genome_buffer,vLB);
      vB        = EPI32_TYPE(_mm_slli)(vB,(1)*8);
      vB        = SI128_TYPE(_mm_or)(vB,vLB);
    }
  }


  /* main loop */
  {
    unsigned int l;
    for (l=0; l<prlength; l++) {

      /* a) get the letters to be compared on the diagonal */
      if (l & 1) {
        VTYPE128 vLB;
        NEXTGENOSEQ_NOSUB_QUAD128(byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                                  byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                                  byte_pos_genome_c,vector_genome_buffer_nbletters_c,
                                  byte_pos_genome_d,vector_genome_buffer_nbletters_d,
                                  vector_genome_buffer,vLB);
        vB  = EPI32_TYPE(_mm_slli)(vB,(1)*8);
        vB  = SI128_TYPE(_mm_or)(vB,vLB);
      } else {
        VTYPE128 vLA;
        NEXTREADSEQ_QUAD128(read,vector_read_buffer_nbletters,
                            vector_read_buffer,vLA);
        vLA = EPI32_TYPE(_mm_slli)(vLA,(3)*8);
        vA  = EPI32_TYPE(_mm_srli)(vA,(1)*8);
        vA  = SI128_TYPE(_mm_or)(vA,vLA);
      }

#ifdef DEBUG
      {
        vector128_t A,B;
        A.v = vA; B.v = vB;
        fprintf(stdout,"[1]\t   A:%.8x,%.8x,%.8x,%.8x\n",(A.ui[3]),(A.ui[2]),(A.ui[1]),(A.ui[0]));
        fprintf(stdout,"[1]\t   B:%.8x,%.8x,%.8x,%.8x\n",(B.ui[3]),(B.ui[2]),(B.ui[1]),(B.ui[0]));
      }
#endif


      VTYPE128 vM;
      /* b) compute the matching score */
      {
        VTYPE128 vM_ab_MatchMask = EPI8_TYPE(_mm_cmpeq)(vA,vB);
        VTYPE128 vM_add = SI128_TYPE(_mm_and)(vM_ab_MatchMask,vMatchS128);
        VTYPE128 vM_sub = SI128_TYPE(_mm_andnot)(vM_ab_MatchMask,vMismatchS128);

#ifdef DEBUG
        {
          vector128_t S_a,S_s;
          S_a.v = vM_add; S_s.v = vM_sub;
          fprintf(stdout,"[1]\t S_a:%.8x,%.8x,%.8x,%.8x\n",(S_a.ui[3]),(S_a.ui[2]),(S_a.ui[1]),(S_a.ui[0]));
          fprintf(stdout,"[1]\t S_s:%.8x,%.8x,%.8x,%.8x\n",(S_s.ui[3]),(S_s.ui[2]),(S_s.ui[1]),(S_s.ui[0]));
        }
#endif


#ifdef DEBUG
        {
          vector128_t M_old_old;
          M_old_old.v = vM_old_old;
          fprintf(stdout,"[1]\t ooM:%.8x,%.8x,%.8x,%.8x\n",(M_old_old.ui[3]),(M_old_old.ui[2]),(M_old_old.ui[1]),(M_old_old.ui[0]));
        }
#endif

        /* same diagonal with M_old_old */
        vM = EPU8_TYPE(_mm_adds)(vM_old_old,vM_add);
        vM = EPU8_TYPE(_mm_subs)(vM,vM_sub);
      }

#ifdef DEBUG
      {
        vector128_t M,M_old,I_old;
        M.v     = vM;
        M_old.v = vM_old;
        I_old.v = vI_old;
        fprintf(stdout,"[1]\t>  M:%.8x,%.8x,%.8x,%.8x\n",(M.ui[3]),(M.ui[2]),(M.ui[1]),(M.ui[0]));
        fprintf(stdout,"[1]\t  oM:%.8x,%.8x,%.8x,%.8x\n",(M_old.ui[3]),(M_old.ui[2]),(M_old.ui[1]),(M_old.ui[0]));
        fprintf(stdout,"[1]\t  oI:%.8x,%.8x,%.8x,%.8x\n",(I_old.ui[3]),(I_old.ui[2]),(I_old.ui[1]),(I_old.ui[0]));
      }
#endif

      VTYPE128 vI;
      {
        /* shift */
        VTYPE128 vM_old_shifted;
        VTYPE128 vI_old_shifted;
        if (l & 1) {
          vM_old_shifted  = EPI32_TYPE(_mm_slli)(vM_old,(1)*8);
          vI_old_shifted  = EPI32_TYPE(_mm_slli)(vI_old,(1)*8);
        } else {
          vM_old_shifted  = EPI32_TYPE(_mm_srli)(vM_old,(1)*8);
          vI_old_shifted  = EPI32_TYPE(_mm_srli)(vI_old,(1)*8);
        }
        VTYPE128 vI_old_merge = EPU8_TYPE(_mm_max)(vI_old,vI_old_shifted);
        VTYPE128 vM_old_merge = EPU8_TYPE(_mm_max)(vM_old,vM_old_shifted);
        vI                    = EPU8_TYPE(_mm_subs)(vI_old_merge,vIndelExtendsS128);
        VTYPE128 vIstart      = EPU8_TYPE(_mm_subs)(vM_old_merge,vIndelOpenS128);
        vI                    = EPU8_TYPE(_mm_max)(vI,vIstart);
        vM                    = EPU8_TYPE(_mm_max)(vM,vI);
      }

#ifdef DEBUG
      {
        vector128_t M,I;
        M.v = vM;
        I.v = vI;
        fprintf(stdout,"[1]\t>  M:%.8x,%.8x,%.8x,%.8x\n",(M.ui[3]),(M.ui[2]),(M.ui[1]),(M.ui[0]));
        fprintf(stdout,"[1]\t>  I:%.8x,%.8x,%.8x,%.8x\n",(I.ui[3]),(I.ui[2]),(I.ui[1]),(I.ui[0]));
      }
#endif

      vM    = SI128_TYPE(_mm_and)(vM,vMsk128[l]);
      vMMax = EPU8_TYPE(_mm_max)(vMMax,vM);

#ifdef DEBUG
      {
        vector128_t Msk,M,Max;
        Msk.v = vMsk128[l];
        M.v   = vM;
        Max.v = vMMax;
        fprintf(stdout,"[1]\t>Msk:%.8x,%.8x,%.8x,%.8x\n",(Msk.ui[3]),(Msk.ui[2]),(Msk.ui[1]),(Msk.ui[0]));
        fprintf(stdout,"[1]\t>  M:%.8x,%.8x,%.8x,%.8x\n",(M.ui[3]  ),(M.ui[2]  ),(M.ui[1]  ),(M.ui[0]  ));
        fprintf(stdout,"[1]\t>Max:%.8x,%.8x,%.8x,%.8x\n",(Max.ui[3]),(Max.ui[2]),(Max.ui[1]),(Max.ui[0]));
      }
#endif

      vM_old_old = vM_old;
      vM_old     = vM;
      vI_old     = vI;

#ifdef DEBUG
      fprintf(stdout,"[1]\t     --------,--------,--------,--------\n");
#endif
    } /* l */
  }
  {
    VTYPE128 vThresholdMask = EPU8_TYPE(_mm_subs)(vMMax,vThreshold128);
    int result = 0;
    int x;
    for(x=0;x<(int)sizeof(VTYPE128)/(int)sizeof(unsigned int);x++) {
      unsigned int u =  ((vector128_t)vThresholdMask).ui[x];
        if (u != 0) {
          result |= (1)<<x;
        }
    }
#ifdef DEBUG
        fprintf(stdout,"[1]\t     ========,========,========,========\n");
#endif
    return result;
  }
}


#endif
















#ifdef __SSE__

/**
 * SSE alignment init read function : modify the read length when needed (but must not be changed too frequently).
 * @param readlength gives the read length (number of nucleotides inside the read)
 */

inline void alignment_sse__setlength_single(unsigned int readlength) {

  /* masking table to keep only good diagonals */
  prlength = (readlength+7) * 2;
  vMsk64[0] = PI8_TYPE(_mm_set)(0xff,0x00,0x00,0x00,
                                0x00,0x00,0x00,0x00);
#ifdef DEBUG
  {
    vector64_t Msk;
    Msk.v = vMsk64[0];
    fprintf(stdout,"[0]\t Msk:%.16llx\n",(Msk.ull[0]));
  }
#endif
  unsigned int l;
  for ( l = 1; l < prlength; l++ ) {
    /* middle mask */
    vMsk64[l] = vMsk64[l - 1];
    if (!(l & 1)) {
      /* mask at the end */
      if (l >= prlength - 7*2) {
        vMsk64[l] = SI64_TYPE(_mm_srli)(vMsk64[l],(1)*8);
      } else {
        /* mask at the beginning */
        if ( l <= 7*2) {
          vMsk64[l] = SI64_TYPE(_mm_or)(vMsk64[l-1],SI64_TYPE(_mm_srli)(vMsk64[l],(1)*8));
        }
      }
    }
#ifdef DEBUG
    {
      vector64_t Msk;
      Msk.v = vMsk64[l];
      fprintf(stdout,"[0]\t Msk:%.16llx\n",(Msk.ull[0]));
    }
#endif
  }/* for l */
}



inline void alignment_sse__setlength_pair(unsigned int readlength) {

  /* masking table to keep only good diagonals */
  prlength = (readlength+3) * 2;
  vMsk64[0] = PI8_TYPE(_mm_set)(0xff,0x00,0x00,0x00,
                                0xff,0x00,0x00,0x00);
#ifdef DEBUG
  {
    vector64_t Msk;
    Msk.v = vMsk64[0];
    fprintf(stdout,"[0]\t Msk:%.8x,%.8x\n",(Msk.ui[1]),(Msk.ui[0]));
  }
#endif
  unsigned int l;
  for ( l = 1; l < prlength; l++ ) {
    /* middle mask */
    vMsk64[l] = vMsk64[l - 1];
    if (!(l & 1)) {
      /* mask at the end */
      if (l >= prlength - 3*2) {
        vMsk64[l] = PI32_TYPE(_mm_srli)(vMsk64[l],(1)*8);
      } else {
        /* mask at the beginning */
        if ( l <= 3*2) {
          vMsk64[l] = SI64_TYPE(_mm_or)(vMsk64[l-1],PI32_TYPE(_mm_srli)(vMsk64[l],(1)*8));
        }
      }
    }
#ifdef DEBUG
    {
      vector64_t Msk;
      Msk.v = vMsk64[l];
      fprintf(stdout,"[0]\t Msk:%.8x,%.8x\n",(Msk.ui[1]),(Msk.ui[0]));
    }
#endif
  }/* for l */
}




/**
 * SSE alignment init function : fix the scoring system and the length of the reads (must be called once before aligning)
 * @param match      inits the match score vector
 * @param mismatch   inits the mismatch penalty vector (positive value only)
 * @param gapopen    inits the gap penalty vector (positive value only)
 * @param gapextends inits the gap penalty vector (positive value only)
 * @param threshold  inits the scoring threshold (positive value only)
 * @param length     fixes the length of the reads that will be treated : this value can be changed @see alignment_sse__setlength_pair
 *        (but must not be changed too frequently).
 */


inline void alignment_sse__init_single(unsigned int match, unsigned int mismatch, unsigned int gapopen, unsigned int gapextends, unsigned int threshold, unsigned int readlength) {

  alignment_sse__compatible_proc();

  /* init some vectors */
  v064             =  PI32_TYPE(_mm_set)(0, 0);
  vBufferMask64    =  PI32_TYPE(_mm_set)(0, 3);
  vThreshold64     =  PI8_TYPE(_mm_set)(threshold, threshold, threshold, threshold,
                                        threshold, threshold, threshold, threshold);
  vIndelOpenS64    =  PI8_TYPE(_mm_set)(gapopen, gapopen, gapopen, gapopen,
                                        gapopen, gapopen, gapopen, gapopen);
  vIndelExtendsS64 =  PI8_TYPE(_mm_set)(gapextends, gapextends, gapextends, gapextends,
                                        gapextends, gapextends, gapextends, gapextends);
  vMatchS64        =  PI8_TYPE(_mm_set)(match, match, match, match,
                                        match, match, match, match);
  vMismatchS64     =  PI8_TYPE(_mm_set)(mismatch, mismatch, mismatch, mismatch,
                                        mismatch, mismatch, mismatch, mismatch);

  alignment_sse__setlength_single(readlength);
}


inline void alignment_sse__init_pair(unsigned int match, unsigned int mismatch, unsigned int gapopen, unsigned int gapextends, unsigned int threshold, unsigned int readlength) {

  alignment_sse__compatible_proc();

  /* init some vectors */
  v064             =  PI32_TYPE(_mm_set)(0, 0);
  vBufferMask64    =  PI32_TYPE(_mm_set)(3, 3);
  vThreshold64     =  PI8_TYPE(_mm_set)(threshold, threshold, threshold, threshold,
                                        threshold, threshold, threshold, threshold);
  vIndelOpenS64    =  PI8_TYPE(_mm_set)(gapopen, gapopen, gapopen, gapopen,
                                        gapopen, gapopen, gapopen, gapopen);
  vIndelExtendsS64 =  PI8_TYPE(_mm_set)(gapextends, gapextends, gapextends, gapextends,
                                        gapextends, gapextends, gapextends, gapextends);
  vMatchS64        =  PI8_TYPE(_mm_set)(match, match, match, match,
                                        match, match, match, match);
  vMismatchS64     =  PI8_TYPE(_mm_set)(mismatch, mismatch, mismatch, mismatch,
                                        mismatch, mismatch, mismatch, mismatch);

  alignment_sse__setlength_pair(readlength);
}




/**
 * SSE alignment align function : does a banded smith-waterman of the given read against two parts of the genome;
 * allows at most 3/4 (4/3) indels on each side.
 * @param genome is the compressed genome (first nucleotide is the lower bit of the first byte)
 * @param pos_genome_a gives the 'A' position (in term of nucleotides) to be aligned with the read (you must substract 3 or 4 potetial indels according to the hit position)
 * @param pos_genome_a gives the 'B' position (in term of nucleotides) to be aligned with the read (you must substract 3 or 4 potetial indels according to the hit position)
 * @param read is the compressed read (first nucleotide is the lower bit of the first byte)
 * @return 0 if none is aligned to reach the given threshold (what shoul happend most of the time),
 *         or a bitmask 1<<(A?0) | 1<<(B?1) if A or B align correctly.
 */


inline int alignment_sse__align_single(unsigned char * genome,
                                       unsigned int * pos_genome,
                                       unsigned char * read) {
  VTYPE64 vA;
  VTYPE64 vB;

  VTYPE64 vMMax;
  VTYPE64 vM_old;
  VTYPE64 vM_old_old;
  VTYPE64 vI_old;

  unsigned int pos_genome_a = pos_genome[0];

  vector64_t     vector_genome_buffer __attribute__ ((aligned (16)));
  unsigned char *             byte_pos_genome = genome + (pos_genome_a >> 2);
  unsigned int                 sub_pos_genome = (pos_genome_a & 3);
  unsigned int vector_genome_buffer_nbletters = 0;

  vector64_t     vector_read_buffer     __attribute__ ((aligned (16)));
  unsigned int vector_read_buffer_nbletters = 0;


  vMMax      = SI64_TYPE(_mm_setzero)();
  vM_old     = SI64_TYPE(_mm_setzero)();
  vM_old_old = SI64_TYPE(_mm_setzero)();
  vI_old     = SI64_TYPE(_mm_setzero)();

  /* init vA : read sequence diagonal */
  {
    vA       = SI64_TYPE(_mm_setzero)();
  }

  /* init vB : genome sequence diagonal */
  {
    int d;
    NEXTGENOSEQ_SINGLE64(sub_pos_genome,byte_pos_genome,vector_genome_buffer_nbletters,vector_genome_buffer,vB);
    for (d=1; d < 8; d++) {
      VTYPE64 vLB;
      NEXTGENOSEQ_NOSUB_NOUP_SINGLE64(vector_genome_buffer_nbletters,vector_genome_buffer,vLB);
      vB        = SI64_TYPE(_mm_slli)(vB,(1)*8);
      vB        = SI64_TYPE(_mm_or)(vB,vLB);
    }
  }


  /* main loop */
  {
    unsigned int l;
    for (l=0; l<prlength; l++) {

      /* a) get the letters to be compared on the diagonal */
      if (l & 1) {
        VTYPE64 vLB;
        NEXTGENOSEQ_NOSUB_SINGLE64(byte_pos_genome,vector_genome_buffer_nbletters,
                                   vector_genome_buffer,vLB);
        vB  = SI64_TYPE(_mm_slli)(vB,(1)*8);
        vB  = SI64_TYPE(_mm_or)(vB,vLB);
      } else {
        VTYPE64 vLA;
        NEXTREADSEQ_SINGLE64(read,vector_read_buffer_nbletters,
                             vector_read_buffer,vLA);
        vLA = SI64_TYPE(_mm_slli)(vLA,(7)*8);
        vA  = SI64_TYPE(_mm_srli)(vA,(1)*8);
        vA  = SI64_TYPE(_mm_or)(vA,vLA);
      }

#ifdef DEBUG
      {
        vector64_t A,B;
        A.v = vA; B.v = vB;
        fprintf(stdout,"[1]\t   A:%.16llx\n",(A.ull[0]));
        fprintf(stdout,"[1]\t   B:%.16llx\n",(B.ull[0]));
      }
#endif


      VTYPE64 vM;
      /* b) compute the matching score */
      {
        VTYPE64 vM_ab_MatchMask = PI8_TYPE(_mm_cmpeq)(vA,vB);
        VTYPE64 vM_add = SI64_TYPE(_mm_and)(vM_ab_MatchMask,vMatchS64);
        VTYPE64 vM_sub = SI64_TYPE(_mm_andnot)(vM_ab_MatchMask,vMismatchS64);

#ifdef DEBUG
        {
          vector64_t S_a,S_s;
          S_a.v = vM_add; S_s.v = vM_sub;
          fprintf(stdout,"[1]\t S_a:%.16llx\n",(S_a.ull[0]));
          fprintf(stdout,"[1]\t S_s:%.16llx\n",(S_s.ull[0]));
        }
#endif


#ifdef DEBUG
        {
          vector64_t M_old_old;
          M_old_old.v = vM_old_old;
          fprintf(stdout,"[1]\t ooM:%.16llx\n",(M_old_old.ull[0]));
        }
#endif

        /* same diagonal with M_old_old */
        vM = PU8_TYPE(_mm_adds)(vM_old_old,vM_add);
        vM = PU8_TYPE(_mm_subs)(vM,vM_sub);
      }

#ifdef DEBUG
      {
        vector64_t M,M_old,I_old;
        M.v     = vM;
        M_old.v = vM_old;
        I_old.v = vI_old;
        fprintf(stdout,"[1]\t>  M:%.16llx\n",(M.ull[0]));
        fprintf(stdout,"[1]\t  oM:%.16llx\n",(M_old.ull[0]));
        fprintf(stdout,"[1]\t  oI:%.16llx\n",(I_old.ull[0]));
      }
#endif

      VTYPE64 vI;
      {
        /* shift */
        VTYPE64 vM_old_shifted;
        VTYPE64 vI_old_shifted;
        if (l & 1) {
          vM_old_shifted  = SI64_TYPE(_mm_slli)(vM_old,(1)*8);
          vI_old_shifted  = SI64_TYPE(_mm_slli)(vI_old,(1)*8);
        } else {
          vM_old_shifted  = SI64_TYPE(_mm_srli)(vM_old,(1)*8);
          vI_old_shifted  = SI64_TYPE(_mm_srli)(vI_old,(1)*8);
        }
        VTYPE64 vI_old_merge  = PU8_TYPE(_mm_max)(vI_old,vI_old_shifted);
        VTYPE64 vM_old_merge  = PU8_TYPE(_mm_max)(vM_old,vM_old_shifted);
        vI                    = PU8_TYPE(_mm_subs)(vI_old_merge,vIndelExtendsS64);
        VTYPE64 vIstart       = PU8_TYPE(_mm_subs)(vM_old_merge,vIndelOpenS64);
        vI                    = PU8_TYPE(_mm_max)(vI,vIstart);
        vM                    = PU8_TYPE(_mm_max)(vM,vI);
      }

#ifdef DEBUG
      {
        vector64_t M,I;
        M.v = vM;
        I.v = vI;
        fprintf(stdout,"[1]\t>  M:%.16llx\n",(M.ull[0]));
        fprintf(stdout,"[1]\t>  I:%.16llx\n",(I.ull[0]));
      }
#endif

      vM    = SI64_TYPE(_mm_and)(vM,vMsk64[l]);
      vMMax = PU8_TYPE(_mm_max)(vMMax,vM);

#ifdef DEBUG
      {
        vector64_t Msk,M,Max;
        Msk.v = vMsk64[l];
        M.v   = vM;
        Max.v = vMMax;
        fprintf(stdout,"[1]\t>Msk:%.16llx\n",(Msk.ull[0]));
        fprintf(stdout,"[1]\t>  M:%.16llx\n",(M.ull[0]));
        fprintf(stdout,"[1]\t>Max:%.16llx\n",(Max.ull[0]));
      }
#endif

      vM_old_old = vM_old;
      vM_old     = vM;
      vI_old     = vI;

#ifdef DEBUG
      fprintf(stdout,"[1]\t     ----------------\n");
#endif
    } /* l */
  }
  {
    VTYPE64 vThresholdMask = PU8_TYPE(_mm_subs)(vMMax,vThreshold64);
    int result = 0;
    int x;
    for(x=0;x<(int)sizeof(VTYPE64)/(int)sizeof(unsigned long long int);x++) {
      unsigned long long int u =  ((vector64_t)vThresholdMask).ull[x];
        if (u != (unsigned long long int) 0) {
          result |= (1)<<x;
        }
    }
#ifdef DEBUG
        fprintf(stdout,"[1]\t     ================\n");
#endif
    return result;
  }
}


inline int alignment_sse__align_pair(unsigned char * genome,
                                     unsigned int * pos_genome,
                                     unsigned char * read) {
  VTYPE64 vA;
  VTYPE64 vB;

  VTYPE64 vMMax;
  VTYPE64 vM_old;
  VTYPE64 vM_old_old;
  VTYPE64 vI_old;

  unsigned int pos_genome_a = pos_genome[0];
  unsigned int pos_genome_b = pos_genome[1];

  vector64_t     vector_genome_buffer __attribute__ ((aligned (16)));
  unsigned char *             byte_pos_genome_a = genome + (pos_genome_a >> 2);
  unsigned int                 sub_pos_genome_a = (pos_genome_a & 3);
  unsigned int vector_genome_buffer_nbletters_a = 0;
  unsigned char *             byte_pos_genome_b = genome + (pos_genome_b >> 2);
  unsigned int                 sub_pos_genome_b = (pos_genome_b & 3);
  unsigned int vector_genome_buffer_nbletters_b = 0;

  vector64_t     vector_read_buffer     __attribute__ ((aligned (16)));
  unsigned int vector_read_buffer_nbletters = 0;


  vMMax      = SI64_TYPE(_mm_setzero)();
  vM_old     = SI64_TYPE(_mm_setzero)();
  vM_old_old = SI64_TYPE(_mm_setzero)();
  vI_old     = SI64_TYPE(_mm_setzero)();

  /* init vA : read sequence diagonal */
  {
    vA       = SI64_TYPE(_mm_setzero)();
  }

  /* init vB : genome sequence diagonal */
  {
    int d;
    NEXTGENOSEQ_PAIR64(sub_pos_genome_a,byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                       sub_pos_genome_b,byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                       vector_genome_buffer,vB);
    for (d=1; d < 4; d++) {
      VTYPE64 vLB;
      /* FIXME gcc 4.3 "crashes" (gcc bug) here when using 
       *  - NEXTGENOSEQ_NOSUB_NOUP_PAIR64
       * so i use now the less opt
       *  - NEXTGENOSEQ_NOSUB_PAIR64
       */
      NEXTGENOSEQ_NOSUB_PAIR64(byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                               byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                               vector_genome_buffer,vLB);
      vB        = PI32_TYPE(_mm_slli)(vB,(1)*8);
      vB        = SI64_TYPE(_mm_or)(vB,vLB);
    }
  }


  /* main loop */
  {
    unsigned int l;
    for (l=0; l<prlength; l++) {

      /* a) get the letters to be compared on the diagonal */
      if (l & 1) {
        VTYPE64 vLB;
        NEXTGENOSEQ_NOSUB_PAIR64(byte_pos_genome_a,vector_genome_buffer_nbletters_a,
                                 byte_pos_genome_b,vector_genome_buffer_nbletters_b,
                                 vector_genome_buffer,vLB);
        vB  = PI32_TYPE(_mm_slli)(vB,(1)*8);
        vB  = SI64_TYPE(_mm_or)(vB,vLB);
      } else {
        VTYPE64 vLA;
        NEXTREADSEQ_PAIR64(read,vector_read_buffer_nbletters,
                           vector_read_buffer,vLA);
        vLA = PI32_TYPE(_mm_slli)(vLA,(3)*8);
        vA  = PI32_TYPE(_mm_srli)(vA,(1)*8);
        vA  = SI64_TYPE(_mm_or)(vA,vLA);
      }

#ifdef DEBUG
      {
        vector64_t A,B;
        A.v = vA; B.v = vB;
        fprintf(stdout,"[1]\t   A:%.8x,%.8x\n",(A.ui[1]),(A.ui[0]));
        fprintf(stdout,"[1]\t   B:%.8x,%.8x\n",(B.ui[1]),(B.ui[0]));
      }
#endif


      VTYPE64 vM;
      /* b) compute the matching score */
      {
        VTYPE64 vM_ab_MatchMask = PI8_TYPE(_mm_cmpeq)(vA,vB);
        VTYPE64 vM_add = SI64_TYPE(_mm_and)(vM_ab_MatchMask,vMatchS64);
        VTYPE64 vM_sub = SI64_TYPE(_mm_andnot)(vM_ab_MatchMask,vMismatchS64);

#ifdef DEBUG
        {
          vector64_t S_a,S_s;
          S_a.v = vM_add; S_s.v = vM_sub;
          fprintf(stdout,"[1]\t S_a:%.8x,%.8x\n",(S_a.ui[1]),(S_a.ui[0]));
          fprintf(stdout,"[1]\t S_s:%.8x,%.8x\n",(S_s.ui[1]),(S_s.ui[0]));
        }
#endif


#ifdef DEBUG
        {
          vector64_t M_old_old;
          M_old_old.v = vM_old_old;
          fprintf(stdout,"[1]\t ooM:%.8x,%.8x\n",(M_old_old.ui[1]),(M_old_old.ui[0]));
        }
#endif

        /* same diagonal with M_old_old */
        vM = PU8_TYPE(_mm_adds)(vM_old_old,vM_add);
        vM = PU8_TYPE(_mm_subs)(vM,vM_sub);
      }

#ifdef DEBUG
      {
        vector64_t M,M_old,I_old;
        M.v     = vM;
        M_old.v = vM_old;
        I_old.v = vI_old;
        fprintf(stdout,"[1]\t>  M:%.8x,%.8x\n",(M.ui[1]),(M.ui[0]));
        fprintf(stdout,"[1]\t  oM:%.8x,%.8x\n",(M_old.ui[1]),(M_old.ui[0]));
        fprintf(stdout,"[1]\t  oI:%.8x,%.8x\n",(I_old.ui[1]),(I_old.ui[0]));
      }
#endif

      VTYPE64 vI;
      {
        /* shift */
        VTYPE64 vM_old_shifted;
        VTYPE64 vI_old_shifted;
        if (l & 1) {
          vM_old_shifted  = PI32_TYPE(_mm_slli)(vM_old,(1)*8);
          vI_old_shifted  = PI32_TYPE(_mm_slli)(vI_old,(1)*8);
        } else {
          vM_old_shifted  = PI32_TYPE(_mm_srli)(vM_old,(1)*8);
          vI_old_shifted  = PI32_TYPE(_mm_srli)(vI_old,(1)*8);
        }
        VTYPE64 vI_old_merge  = PU8_TYPE(_mm_max)(vI_old,vI_old_shifted);
        VTYPE64 vM_old_merge  = PU8_TYPE(_mm_max)(vM_old,vM_old_shifted);
        vI                    = PU8_TYPE(_mm_subs)(vI_old_merge,vIndelExtendsS64);
        VTYPE64 vIstart       = PU8_TYPE(_mm_subs)(vM_old_merge,vIndelOpenS64);
        vI                    = PU8_TYPE(_mm_max)(vI,vIstart);
        vM                    = PU8_TYPE(_mm_max)(vM,vI);
      }

#ifdef DEBUG
      {
        vector64_t M,I;
        M.v = vM;
        I.v = vI;
        fprintf(stdout,"[1]\t>  M:%.8x,%.8x\n",(M.ui[1]),(M.ui[0]));
        fprintf(stdout,"[1]\t>  I:%.8x,%.8x\n",(I.ui[1]),(I.ui[0]));
      }
#endif

      vM    = SI64_TYPE(_mm_and)(vM,vMsk64[l]);
      vMMax = PU8_TYPE(_mm_max)(vMMax,vM);

#ifdef DEBUG
      {
        vector64_t Msk,M,Max;
        Msk.v = vMsk64[l];
        M.v   = vM;
        Max.v = vMMax;
        fprintf(stdout,"[1]\t>Msk:%.8x,%.8x\n",(Msk.ui[1]),(Msk.ui[0]));
        fprintf(stdout,"[1]\t>  M:%.8x,%.8x\n",(M.ui[1]),(M.ui[0]));
        fprintf(stdout,"[1]\t>Max:%.8x,%.8x\n",(Max.ui[1]),(Max.ui[0]));
      }
#endif

      vM_old_old = vM_old;
      vM_old     = vM;
      vI_old     = vI;

#ifdef DEBUG
      fprintf(stdout,"[1]\t     --------,--------\n");
#endif
    } /* l */
  }
  {
    VTYPE64 vThresholdMask = PU8_TYPE(_mm_subs)(vMMax,vThreshold64);
    int result = 0;
    int x;
    for(x=0;x<(int)sizeof(VTYPE64)/(int)sizeof(unsigned int);x++) {
      unsigned int u =  ((vector64_t)vThresholdMask).ui[x];
        if (u != 0) {
          result |= (1)<<x;
        }
    }
#ifdef DEBUG
        fprintf(stdout,"[1]\t     ========,========\n");
#endif
    return result;
  }
}

#endif
