#ifndef READS_DATA_H_
#define READS_DATA_H_

#include "load_data.h"
#include "read_quality.h"

#define DEFAULT_READ_LEN 35
#define MAX_HEADER_LEN   64

#define TITLE_LINE_MARKER "# Title: "
#define TITLE_LINE_MARKER_LEN 9

/**
 * @struct ReadDataType
 * data stored for one read
 */

typedef struct ReadDataType {
  //int        id;
  /** header descriptor in the "csfasta/fastq" file */
  char      *info;
  /** first color (synch color, not used in the alignment process) */
  CODE_TYPE  first_base;
  /** first quality (associated to synch color) */
  QUAL_TYPE  first_qual;
  /** color space read sequence */
  CODE_TYPE *sequence;
  /** color space read quality */
  QUAL_TYPE *quality;
  /** read already processed */
  char disabled;
} ReadDataType;


/**
 * @struct ReadDataType
 * data stored for a set of reads
 * @see ReadDataType
 */

typedef struct ReadsDBType {
  /** descriptor name of the set of reads */
  char* name;
  /** length of the reads (all must have the same !!) */
  int read_len;
  /** number of reads */
  int size;
  /** array storing the set of reads */
  ReadDataType* reads;
} ReadsDBType;

/**
 * Creates an empty read structure
 */
ReadDataType* read__create(int read_len);

/**
 * Allocate memory for all the read data
 */
int read__alloc_vectors(ReadDataType* read, int read_len);

/**
 * Reverse-complement a read
 */
inline void read__reverse(const ReadDataType* src, ReadDataType* dest, int len);

/**
 * Destroys a read database
 */
void read__destroy(ReadDataType* read);

/**
 * Loads a reads database
 */
int load_reads_db(const char* reads_filename, const char* qual_filename, ReadsDBType* db);
/**
 * Destroys a read database
 */
void clear_reads_db(ReadsDBType* db);

#endif /* READS_DATA_H_ */
