#include "run.h"
#include <time.h>
#ifdef _OPENMP
#include <omp.h>
#endif

#define USE_EXTERNAL_HIT_POINTER

#include "stats.h"

#define MAX_MULTIPLE_HITS 4

#ifdef __SSE2__
const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {4, 4, 4, 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, unsigned int *, unsigned char *) = {
    &alignment_sse2__align_quad,
    &alignment_sse2__align_quad,
    &alignment_sse2__align_quad,
    &alignment_sse2__align_quad,
    &alignment_sse2__align_pair,
    &alignment_sse2__align_pair,
    &alignment_sse2__align_pair,
    &alignment_sse2__align_pair,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single,
    &alignment_sse2__align_single
    };
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
    &alignment_sse2__init_quad,
    &alignment_sse2__init_quad,
    &alignment_sse2__init_quad,
    &alignment_sse2__init_quad,
    &alignment_sse2__init_pair,
    &alignment_sse2__init_pair,
    &alignment_sse2__init_pair,
    &alignment_sse2__init_pair,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single,
    &alignment_sse2__init_single
    };
#else
#ifdef __SSE__
const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {2, 2, 2, 2, 1, 1, 1, 1};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, unsigned int *, unsigned char *) = {
    &alignment_sse__align_pair,
    &alignment_sse__align_pair,
    &alignment_sse__align_pair,
    &alignment_sse__align_pair,
    &alignment_sse__align_single,
    &alignment_sse__align_single,
    &alignment_sse__align_single,
    &alignment_sse__align_single
    };
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
    &alignment_sse__init_pair,
    &alignment_sse__init_pair,
    &alignment_sse__init_pair,
    &alignment_sse__init_pair,
    &alignment_sse__init_single,
    &alignment_sse__init_single,
    &alignment_sse__init_single,
    &alignment_sse__init_single
    };
#else

void fake_init (unsigned int match,
        unsigned int mismatch,
        unsigned int gapopen,
        unsigned int gapextends,
        unsigned int threshold,
        unsigned int readlength) {}
int fake_align (unsigned char * genome,
        unsigned int * pos_genome,
        unsigned char * read) { return 1; /* 0x01 */}

const int simd_mul[INDEL_DATA_VECTOR_SIZE] = {1, 1, 1, 1, 1, 1, 1, 1};
void(* const simd_init_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned int, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int) = {
    &fake_init,
    &fake_init,
    &fake_init,
    &fake_init,
    &fake_init,
    &fake_init,
    &fake_init,
    &fake_init
};
int(* const simd_fct_table[INDEL_DATA_VECTOR_SIZE])(unsigned char *, unsigned int *, unsigned char *) = {
    &fake_align,
    &fake_align,
    &fake_align,
    &fake_align,
    &fake_align,
    &fake_align,
    &fake_align,
    &fake_align
};

#endif
#endif

int SHOW_TRACEBACK_PATTERNS = 0;
int LIST_UNMAPPED_READS = 0;

//#define DEBUG__OUTPUT
//#define __DONT__MAP__

/**
 * Statistics for finding the efficiency of the filters
 */

/* #define ENABLE_STATS */

#ifdef ENABLE_STATS
#define IDX_PROCESSED_READS 0
#define IDX_SEED_HITS 1
#define IDX_SIMD_PASSED 2
#define IDX_ALGO_PASSED 3
#define STATS_FIELDS_COUNT 4


int FILTER_STATS[4][4] = {
    {0,0,0,0},
    {0,0,0,0},
    {0,0,0,0},
    {0,0,0,0}
};

int *stats__seed_idx;

void stats__init() {
#ifdef _OPENMP
  SAFE_FAILURE__ALLOC(stats__seed_idx, MAX_THREADS, int);
#else
  SAFE_FAILURE__ALLOC(stats__seed_idx, 1, int);
#endif
}

#ifdef _OPENMP
#define stats__set_seed_idx(seed_idx) {stats__seed_idx[_ogtn] = seed_idx;}
#else
#define stats__set_seed_idx(seed_idx) {stats__seed_idx[0] = seed_idx;}
#endif


#ifdef _OPENMP
#define stats__inc(field_idx,count)  FILTER_STATS[stats__seed_idx[_ogtn]][field_idx] += count;
#else
#define stats__inc(field_idx,count)  FILTER_STATS[stats__seed_idx[0]][field_idx] += count;
#endif


void stats__destroy() {
  free(stats__seed_idx);
}

void stats__display(IndexType* index, int* stats) {
#ifdef DEBUG_INDEX
    INFO__(("%s:\n  REAL HITS:      %10d",              seed__to_string(index->seed), index__get_hits_count(index)));
    INFO__((     "  EVALUATED HITS: %10d (%3.2lf%%)",   stats[IDX_SEED_HITS],   stats[IDX_SEED_HITS]   * 100.0 / index->hits_count     ));
#else
    INFO__((     "  EVALUATED HITS: %10d",   stats[IDX_SEED_HITS] ));
#endif
    INFO__((     "  SIMD FILTER:    %10d (%3.2lf%%)",   stats[IDX_SIMD_PASSED], stats[IDX_SIMD_PASSED] * 100.0 / stats[IDX_SEED_HITS]  ));
    INFO__((     "  FULL ALIGNMENT: %10d (%3.2lf%%)\n", stats[IDX_ALGO_PASSED], stats[IDX_ALGO_PASSED] * 100.0 / stats[IDX_SIMD_PASSED]));
}

#endif

/**
 * Perform an alignment between two sequences
 * @param read_str First sequence (color code)
 * @param ref_str Second sequence (color code)
 * @param read quality values for the first sequence
 * @param match
 * @param mismatch
 * @param gap_open
 * @param gap_extend
 * @param allowed_indels
 * @param output The output file descriptor
 */
int single_alignment(char* read_str, char* ref_str, char* qual_str,
    ScoreType match, ScoreType mismatch, ScoreType gap_open, ScoreType gap_extend, short allowed_indels,
    FILE* output) {
  CODE_TYPE* read = string_to_code(read_str);
  CODE_TYPE* ref  = string_to_code(ref_str);
  int read_len = strlen(read_str);
  short *quality = NULL;
  ScoreType best;
  AlignmentType* alignment;

  VERBOSITY = VERBOSITY_MODERATE; // to allow the displayed alignment to be built

  if (qual_str) {
    quality = parse_integer_sequence(qual_str, read_len);
  }

  alignment = alignment__create1(read, quality, strlen(read_str), ref, strlen(ref_str), allowed_indels);
  alignment__set_params(alignment, match, mismatch, gap_open, gap_extend);

  best = alignment__align(alignment, MIN_SCORE);

  fprintf(output, "\n%s\n" , alignment__traceback(alignment));
  fprintf(output, "Score: %d\n\n", best);
  alignment__destroy(alignment);
  free(read);
  free(ref);
  if (quality) {
    free(quality);
  }

  return best;
}

/*----------------------------------------------8<--------------------------------------------------------
 * For multithreaded execution via openmp, some data structures need to be cloned for each thread
 */
#ifdef _OPENMP

int MAX_THREADS  = DEFAULT_MAX_THREADS;

#define QUALITY_MASK__DECLARE        unsigned int ** quality_mask;
#define QUALITY_MASK__INIT       \
  SAFE_FAILURE__ALLOC(quality_mask, MAX_THREADS, unsigned int*);\
  for (t = 0; t < MAX_THREADS; ++t) {\
    SAFE_FAILURE__ALLOC(quality_mask[t], QUALITY_MASK_SIZE, unsigned int); \
    memset(quality_mask[t], 0x00, QUALITY_MASK_SIZE * sizeof(unsigned int)); \
    }
#define QUALITY_MASK_REF             quality_mask[_ogtn]
#define QUALITY_MASK__DESTROY                \
  for (t = 0; t < MAX_THREADS; ++t) {\
    free(quality_mask[t]);\
  }\
  free(quality_mask);


/*Alignment structures for each process */

#define ALIGNMENT__DECLARE(alignment) int t; AlignmentType* alignment;

#define ALIGNMENT__INIT(alignment, read_len, ref_len, allowed_indels, match, mismatch, gap_open, gap_extend) \
SAFE_FAILURE__ALLOC(alignment, MAX_THREADS, AlignmentType); \
for (t = 0; t < MAX_THREADS; ++t) { \
  alignment__init(&alignment[t], read_len, ref_len, allowed_indels); \
  alignment__set_params(&alignment[t], match, mismatch, gap_open, gap_extend); \
}

#define ALIGNMENT__RESET_READ(alignment, read, quality)\
  alignment__reset_with_compressed_read(&alignment[_ogtn], read, quality);\

#define ALIGNMENT__REF(alignment) &alignment[_ogtn]

#define ALIGNMENT__DESTROY(alignment) \
for (t = 0; t < MAX_THREADS; ++t) { \
  alignment__destroy(&alignment[t]);\
}\
free(alignment);


/* Filters for each process */

#define FILTER__DECLARE(filter) RegionFilterType* filter;

#define FILTER__INIT(filter, ref_db, ref_window)\
SAFE_FAILURE__ALLOC(filter, MAX_THREADS, RegionFilterType); \
for (t = 0; t < MAX_THREADS; ++t) { \
  reg_filter__init(&filter[t], ref_db, ref_window);\
}

#define FILTER__REF(filter) &filter[_ogtn]

#define FILTER__DESTROY(filter) \
for (t = 0; t < MAX_THREADS; ++t) { \
  reg_filter__destroy(&filter[t]); \
}\
free(filter);


/* Reverse complementary data */
#define REV_SEQ__DECLARE(tmp_sequence)  ReadDataType *tmp_sequence;

#define REV_SEQ__ALLOC(tmp_sequence, len) \
SAFE_FAILURE__ALLOC(tmp_sequence, MAX_THREADS, ReadDataType); \
for (t = 0; t < MAX_THREADS; ++t){\
  SAFE_FAILURE__ALLOC( tmp_sequence[t].sequence, len, CODE_TYPE); \
    tmp_sequence[t].quality = NULL;\
}

#define REV_SEQ__REF(tmp_sequence)     &tmp_sequence[_ogtn]
#define REV_SEQ__SEQ_REF(tmp_sequence)  tmp_sequence[_ogtn].sequence
#define REV_SEQ__QUAL_REF(tmp_sequence) tmp_sequence[_ogtn].quality

#define REV_SEQ__DESTROY(tmp_sequence) \
for (t = 0; t < MAX_THREADS; ++t){\
    free(tmp_sequence[t].sequence);\
    if (tmp_sequence[t].quality) { \
      free(tmp_sequence[t].quality); \
    }\
}\
free(tmp_sequence);

#else

#define QUALITY_MASK__DECLARE        unsigned int quality_mask[QUALITY_MASK_SIZE];
#define QUALITY_MASK__INIT           memset(quality_mask, 0x00, QUALITY_MASK_SIZE * sizeof(unsigned int));
#define QUALITY_MASK_REF             quality_mask
#define QUALITY_MASK__DESTROY


/* Alignment Structure */

#define ALIGNMENT__DECLARE(alignment) AlignmentType alignment;

#define ALIGNMENT__INIT(alignment, read_len, ref_len, allowed_indels, match, mismatch, gap_open, gap_extend) \
  alignment__init(&alignment, read_len, ref_len, allowed_indels); \
  alignment__set_params(&alignment, match, mismatch, gap_open, gap_extend);
#define ALIGNMENT__RESET_READ(alignment, read, quality)\
  alignment__reset_with_compressed_read(&alignment, read, quality);
#define ALIGNMENT__REF(alignment) &alignment
#define ALIGNMENT__DESTROY(alignment) alignment__destroy(&alignment);

/* Filter */

#define FILTER__DECLARE(filter) RegionFilterType filter;
#define FILTER__INIT(filter, ref_db, ref_window) reg_filter__init(&filter, ref_db, ref_window);
#define FILTER__REF(filter) &filter
#define FILTER__DESTROY(filter) reg_filter__destroy(&filter);

/* Reverse complementary data */
#define REV_SEQ__DECLARE(tmp_sequence)  ReadDataType tmp_sequence;
#define REV_SEQ__ALLOC(tmp_sequence, len) SAFE_FAILURE__ALLOC(tmp_sequence.sequence, len, CODE_TYPE); tmp_sequence.quality = NULL;
#define REV_SEQ__REF(tmp_sequence)  &tmp_sequence
#define REV_SEQ__SEQ_REF(tmp_sequence)  tmp_sequence.sequence
#define REV_SEQ__QUAL_REF(tmp_sequence) tmp_sequence.quality
#define REV_SEQ__DESTROY(tmp_sequence) free(tmp_sequence.sequence); if (tmp_sequence.quality) {free(tmp_sequence.quality);}

#endif


/*---------------------------------------------->8--------------------------------------------------------*/

/**
 * Find and register all the hits of applying a certain seed at a given position in a read
 * @param read
 * @param ri The read's index (for updating the map
 * @param j The position where the seed should be applied
 * @param filter Constrains references to the seed, the index AND the reference database
 * @param alignment The alignment data structure
 * @param map the read map to update with good alignments
 */
static inline void process_read_with_seed__m(
               CODE_TYPE* read_sequence, const int ri, const int j,
               RegionFilterType *filter, IndexType* index,
               AlignmentType* alignment, HitMapType* map,
               const int alignment_sense) {
  int k, h = 0, updates = 0, i, p;
  int I = alignment->params.allowed_indels;


  CODE_TYPE* ref_seq  = ((ReferenceDBType*)index->db)->sequence;
  int ref_size = ((ReferenceDBType*)index->db)->size;
  ScoreType score, score_threshold = min_accepted_score;

  int hits[MAX_MULTIPLE_HITS];
  unsigned int ref_window[MAX_MULTIPLE_HITS];

#if (!defined USE_EXTERNAL_HIT_POINTER) || (defined ENABLE_STATS)
#ifdef _OPENMP
  const int _ogtn = omp_get_thread_num();
#else
  const int _ogtn = 0;
#endif
#endif

  /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
  stats__inc(IDX_PROCESSED_READS, 1);
#endif
  // get the resulted code k
  k = seed__apply_to_compressed(index->seed, (read_sequence + COMPRESSED_IDX(j)), COMPRESSED_OFFSET(j));

#ifdef USE_EXTERNAL_HIT_POINTER
  int hit_pointer = -1;
#else
  index__reset_crt_hit(index, _ogtn);
#endif

#define IS_HIT_POSITION_VIABLE(h) ((ref_size - (h) - I) >= (alignment->read_len - j))

  // search it in the index i
  while (h != -1) {
#ifdef USE_EXTERNAL_HIT_POINTER
    h = reg_filter__extern_next_hits(filter, index, k, hits, simd_mul[I], &hit_pointer);
#else
    h = reg_filter__next_hits(filter, index, k, hits, simd_mul[I], _ogtn);
#endif
    // no hit at all? exit loop
    if (hits[0] == -1 || !IS_HIT_POSITION_VIABLE(hits[0])) {
      break;
    }
    for (i = 0; i < simd_mul[I]; ++i) {
      if ((i == 0 || hits[i] != hits[i-1]) && IS_HIT_POSITION_VIABLE(hits[i])) {
  ref_window[i] = max(hits[i] - j - I, 0);
  /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
  stats__inc(IDX_SEED_HITS, 1);
#endif
  VERB_FILTER(VERBOSITY_ANNOYING, INFO__(("\nRead %d[%d]: HIT Reference@%d", ri, j, hits[i])););
      } else {
  ref_window[i] = ref_window[i-1];
      }
    }
    i = simd_fct_table[I](ref_seq, ref_window, read_sequence);
    p = 0;
    while (i && IS_HIT_POSITION_VIABLE(hits[p])) {
      if((i & 1) && (p == 0 || ref_window[p] != ref_window[p-1])) {
  VERB_FILTER(VERBOSITY_ANNOYING, INFO__(("\nRead %d PASSED SIMD FILTER Reference@%d", ri, ref_window[p])););

  /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
  stats__inc(IDX_SIMD_PASSED, 1);
#endif
  alignment__reset_with_compressed_ref(alignment, ref_seq, ref_window[p]);
  //If the hit is not in a region that has already been checked, and the alignment score is acceptable, memorize it
  if ((score = alignment__align(alignment, score_threshold)) > score_threshold) {

    /* @filter_efficiency_statistics */
#ifdef ENABLE_STATS
    stats__inc(IDX_ALGO_PASSED, 1);
#endif
    alignment__traceback(alignment);
    // mask the aligned zone
    reg_filter__mask(filter, alignment, ref_window[p]);
    // display the alignment
    VERB_FILTER(VERBOSITY_HIGH, INFO__(("\nReference@%d, Read %d\n%sScore: %d\n", ref_window[p] + alignment->best_j0, ri, alignment->to_display, score)););
    // memorize in a map
#ifdef _OPENMP
#pragma omp critical(mask)
#endif
    {
      updates += (hit_map__update(map, ri, ref_window[p], alignment, alignment_sense) >= 0) ? 1 : 0;
    }
    if (map->map[ri][MAP_DETAIL_SIZE-1].score > score_threshold) {
      score_threshold = map->map[ri][MAP_DETAIL_SIZE-1].score;
    }
  }
      }
      i >>= 1;
      ++p;
    }
  }
}
/**
 * Find and register all the hits of applying a certain seed at a given position in a read
 * @param read
 * @param ri The read's index (for updating the map
 * @param j The position where the seed should be applied
 * @param filter Constrains references to the seed, the index AND the reference database
 * @param alignment The alignment data structure
 * @param map the read map to update with good alignments
 */

static inline void process_read_with_seed(
    CODE_TYPE* read_sequence, const int ri, const int j,
    RegionFilterType *filter, IndexType* index,
    AlignmentType* alignment, HitMapType* map,
    const int alignment_sense) {
  process_read_with_seed__m(read_sequence,ri,j,filter,index,alignment,map,alignment_sense);
}

/**
 * Reads mapped on a reference
 * @param reads_filename
 * @param qual_filename
 * @param ref_filename
 * @param seed_list
 * @param match
 * @param mismatch
 * @param gap_open
 * @param gap_extend
 * @param allowd_indels
 * @param output
 * @param seq_output
 */
int reads_against_reference(const char* reads_filename, const char* qual_filename, const char* ref_filename,
    const char* seedslist,
    ScoreType match, ScoreType mismatch, ScoreType gap_open, ScoreType gap_extend, short allowed_indels,
    FILE* output, FILE* seq_output) {
  ReadsDBType       reads_db;
  ReferenceDBType*  ref_dbs;
  ReferenceDBType*  ref_db;
  SeedType **       seeds;
  IndexType**       ref_index;
  int refs_count, seed_count, ref_window_len, ri, p_ri, si, rfi, min_seed_len;

#ifdef _OPENMP
  omp_set_num_threads(MAX_THREADS);
#endif

  ALIGNMENT__DECLARE(alignment);
  FILTER__DECLARE(filter);

  //CODE_TYPE *tmp_sequence;
  REV_SEQ__DECLARE(tmp_sequence);

  QUALITY_MASK__DECLARE

  HitMapType* map;
  GenomeMapType* genome_map;

  time_t start_time = time(NULL), crt_time, crt_time2;

#ifdef ENABLE_STATS
  stats__init();
#endif

  /* create seeds */
  seed_count = seed__parse_list(seedslist, &seeds);
  /* verify seed validity */
  if (seed_count == RETURN_INPUT_ERR) {
    ERROR__(("The seed pattern is invalid: %s. Expected: {0,1}+, at least one '1', or {-,#)+, at least one '#'.", seedslist));
    exit (RETURN_INPUT_ERR);
  } else if (seed_count <= 0) {
    ERROR__(("The seed input is invalid: %s. Please provide at least one seed.", seedslist));
    exit (RETURN_INPUT_ERR);
  }
  VERB_FILTER(VERBOSITY_MODERATE, INFO__(("\nThe following %d seed%s will be used:\n", seed_count, (seed_count > 1 ? "s" : ""))); for (si = 0; si < seed_count; ++si) { seed__display(seeds[si]); });

  // load the two databases
  message(VERBOSITY_MODERATE, "\nLoading reads database...");
  crt_time = time(NULL);
  if (load_reads_db    (reads_filename, qual_filename, &reads_db) <= 0) {
  ERROR__(("Reads could not be loaded. Exiting."));
    return RETURN_INPUT_ERR;
  }

  //  message(VERBOSITY_MODERATE, "Initializing map...");
  //  crt_time = time(NULL);
  map = hit_map__create(reads_db.size, allowed_indels);
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loaded %d reads in %ld seconds.\n\n", reads_db.size, time(NULL) - crt_time);)



  message(VERBOSITY_MODERATE, "Loading reference database...");
  crt_time = time(NULL);
  if ((refs_count = load_reference_db(ref_filename, &ref_dbs)) <= 0) {
    ERROR__(("Reference could not be loaded. Exiting."));
    return RETURN_INPUT_ERR;
  }
  VERB_FILTER(VERBOSITY_MODERATE, printf("Loaded %d reference sequence%s in %ld seconds.\n\n", refs_count, (refs_count == 1 ? "" : "s"), time(NULL) - crt_time);)

  // create alignment data
  ref_window_len = reads_db.read_len + (allowed_indels<<1);
  // SIMD filter initialization
  simd_init_fct_table[allowed_indels](abs(match),abs(mismatch),abs(gap_open),abs(gap_extend),min_accepted_score_simd_filter, reads_db.read_len);

  ALIGNMENT__INIT(alignment, reads_db.read_len, ref_window_len, allowed_indels, match, mismatch, gap_open, gap_extend);
  REV_SEQ__ALLOC(tmp_sequence, reads_db.read_len);
  /* min seed length, for masking... */
  min_seed_len = reads_db.read_len;
  for (rfi = 0; rfi < seed_count; ++rfi) {
    if (min_seed_len > seeds[rfi]->length) {
      min_seed_len = seeds[rfi]->length;
    }
  }
  /* Create indexes */
  SAFE_FAILURE__ALLOC(ref_index, seed_count, IndexType*);

  QUALITY_MASK__INIT

  /**********************************************
   * Process each reference sequence
   */
  for (rfi = 0; rfi < refs_count; ++rfi) {
    ref_db = &ref_dbs[rfi];
    VERB_FILTER(VERBOSITY_MODERATE, INFO__(("\n\nProcessing reference sequence %s (%d of %d, %d colors)...\n", ref_db->name, rfi + 1, refs_count, ref_db->size));)
    // create index
    message(VERBOSITY_MODERATE, "Creating reference index...");
    crt_time = time(NULL);
#ifdef _OPENMP
#pragma omp parallel for shared(ref_index, ref_window_len, ref_db, seeds)
#endif
    for (si = 0; si < seed_count; ++si) {
      ref_index[si] = index__build_reference(ref_db, seeds[si]);
    }
    VERB_FILTER(VERBOSITY_MODERATE, printf("Created %d indexes in %ld seconds.\n\n", seed_count, time(NULL) - crt_time);)

    // create filter
    FILTER__INIT(filter, ref_db, (ref_window_len - min_seed_len));

    // On each read, pas each seed, find the hits, and align
    message(VERBOSITY_MODERATE, "Start search...");

    crt_time = time(NULL);
    if (ALIGNMENT_SENSE & ALIGNMENT_SENSE_FORWARD) {
      message(VERBOSITY_MODERATE, "Forward...");
      p_ri = 0;

#ifdef _OPENMP
#pragma omp parallel for shared(p_ri, reads_db, ref_index, seeds, quality_mask) schedule(dynamic,1000)
#endif
      for (ri = 0; ri < reads_db.size; ++ri) {
        int i = 0, j = 0, read_seed_positions;
#ifdef _OPENMP
	int _ogtn = omp_get_thread_num();
#endif

        if (reads_db.reads[ri].disabled) {
#ifdef _OPENMP
#pragma omp critical(progress)
#endif
          p_ri++;
          continue;
        }
        //reseting region filter...
        reg_filter__reset(FILTER__REF(filter));
        // set the read ri
        ALIGNMENT__RESET_READ(alignment,  reads_db.reads[ri].sequence, reads_db.reads[ri].quality);
        /* [NEW] compute read quality mask */
        if (reads_db.reads[ri].quality) {
          quality__build_mask(reads_db.reads[ri].quality, reads_db.read_len, QUALITY_MASK_REF);
        }
        // apply each seed i
        for (i = 0; i < seed_count; ++i) {

          /* @filter_efficiency_statistics */
          #ifdef ENABLE_STATS
          stats__set_seed_idx(i);
          #endif

          read_seed_positions = reads_db.read_len - seeds[i]->length;
          // on each position j
          if (seeds[i]->positions) {
          // if the seed comes with specific positions, use those
            for (j = 0; j < seeds[i]->positions_count; ++j) {
              // better check if it can be applied, to prevent any overflows
              if (seeds[i]->positions[j] <= read_seed_positions) {
                process_read_with_seed((reads_db.reads[ri]).sequence, ri, seeds[i]->positions[j], FILTER__REF(filter), ref_index[i], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_FORWARD);
              }
            }
          } else {
            // else, apply it everywhere, filtering according to good quality zones [NEW]
            for (j = 0; j <= read_seed_positions; ++j) {
              if (reads_db.reads[ri].quality == NULL || quality__accept_seed_relaxed(QUALITY_MASK_REF, reads_db.read_len, seeds[i], j, MIN_TRUSTED_COLORS)) {
                process_read_with_seed((reads_db.reads[ri]).sequence, ri, j, FILTER__REF(filter), ref_index[i], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_FORWARD);
              }
            }
          }
        }
#ifdef _OPENMP
#pragma omp critical(progress)
#endif
        {
          VERB_FILTER(VERBOSITY_MODERATE, display_progress(p_ri, reads_db.size, map->mapped))
          p_ri++;
        }
      }
      VERB_FILTER(VERBOSITY_NONE, printf("\nForward completed in %ld seconds, %d of %d reads aligned.\n", time(NULL) - crt_time, map->mapped, map->size))
    }

    crt_time2 = time(NULL);
    if (ALIGNMENT_SENSE & ALIGNMENT_SENSE_REVERSE) {
      message(VERBOSITY_MODERATE, "Reverse complementary...");
      p_ri = 0;

#ifdef _OPENMP
#pragma omp parallel for shared(p_ri, reads_db, ref_index, seeds, quality_mask) schedule(dynamic,1000)
#endif
      for (ri = 0; ri < reads_db.size; ++ri) {
	int i = 0, j = 0, read_seed_positions;
#ifdef _OPENMP
	int _ogtn = omp_get_thread_num();
#endif

        if (reads_db.reads[ri].disabled) {
#ifdef _OPENMP
#pragma omp critical(progress)
#endif
          p_ri++;
          continue;
        }
        //reseting region filter...
        reg_filter__reset(FILTER__REF(filter));
        // set the read ri
        read__reverse(&reads_db.reads[ri], REV_SEQ__REF(tmp_sequence), reads_db.read_len);

        ALIGNMENT__RESET_READ(alignment, REV_SEQ__SEQ_REF(tmp_sequence), REV_SEQ__QUAL_REF(tmp_sequence));
        /* [NEW] compute read quality mask */
        if (reads_db.reads[ri].quality) {
          quality__build_mask(REV_SEQ__QUAL_REF(tmp_sequence), reads_db.read_len, QUALITY_MASK_REF);
        }
          // apply each seed i
        for (i = 0; i < seed_count; ++i) {
          /* @filter_efficiency_statistics */
          #ifdef ENABLE_STATS
            stats__set_seed_idx(i);
          #endif

          read_seed_positions = reads_db.read_len - seeds[i]->length;
          // on each position j
          if (seeds[i]->positions) {
            // if the seed comes with specific positions, use those
            for (j = 0; j < seeds[i]->positions_count; ++j) {
              // better check if it can be applied, to prevent any overflows
              if (seeds[i]->positions[j] <= read_seed_positions) {
                process_read_with_seed(REV_SEQ__SEQ_REF(tmp_sequence), ri, read_seed_positions - seeds[i]->positions[j], FILTER__REF(filter), ref_index[i], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_REVERSE);
              }
            }
          } else {
            // else, apply it everywhere
            for (j = 0; j <= read_seed_positions; ++j) {
              if (reads_db.reads[ri].quality == NULL || quality__accept_seed_relaxed(QUALITY_MASK_REF, reads_db.read_len, seeds[i], j, MIN_TRUSTED_COLORS)) {
                process_read_with_seed(REV_SEQ__SEQ_REF(tmp_sequence), ri, j, FILTER__REF(filter), ref_index[i], ALIGNMENT__REF(alignment), map, ALIGNMENT_SENSE_REVERSE);
              }
            }
          }
        }
        if (VERBOSITY <= VERBOSITY_MODERATE) {
          /* Do not display this progress bar if other details (such as specific local alignments) are also displayed. */
#ifdef _OPENMP
#pragma omp critical(progress)
#endif
        {
          VERB_FILTER(VERBOSITY_MODERATE, display_progress(p_ri, reads_db.size, map->mapped))
          p_ri++;
        }
        }
      }
      VERB_FILTER(VERBOSITY_NONE, printf("\nReverse complementary completed in %ld seconds.\n", time(NULL) - crt_time2))
    }
    VERB_FILTER(VERBOSITY_NONE, printf("\nAligned %d of %d reads (%5.2f%%) in %ld seconds. Mapping...\n", map->mapped, map->size, (map->mapped*100.0/map->size), time(NULL) - crt_time))

    crt_time = time(NULL);

    /* Erase the indexes to make room for the mapping structures */

    for (si = 0; si < seed_count; ++si) {
      index__destroy(ref_index[si]);
    }

#ifndef __DONT__MAP__

    /* Generate mapping */
    genome_map = genome_map__create(map, ref_db, &reads_db);
    genome_map__build(genome_map);
    /* It may be useful to see what the alignments look like */
    if (SHOW_TRACEBACK_PATTERNS) {
      display_tracebacks(genome_map);
    }
    /* It may be useful to see which reads didn't make it */
    if (LIST_UNMAPPED_READS) {
      list_unmapped_reads(&reads_db);
      list_unmapped_reads_translated(&reads_db);
    }

    /* Generate output */
    VERB_FILTER(VERBOSITY_NONE, printf("\nMap built in %ld seconds. Generating output...\n", time(NULL) - crt_time);)
#ifdef DEBUG__OUTPUT
    genome_map__generate_output(genome_map, output, seq_output);
#else
    genome_map__generate_SAM_output(genome_map, output);
#endif
    /* More cleanup cleanup */
    genome_map__destroy(genome_map);
    hit_map__clean(map);
    clear_reference_db(ref_db);
    VERB_FILTER(VERBOSITY_MODERATE, printf("Reference sequence %d done in %ld seconds.\n\n", rfi+1, time(NULL) - crt_time);)
#endif /*__DONT__MAP__*/
  }
  // more cleanup
  hit_map__destroy(map);
  free(map);
  clear_reads_db(&reads_db);
  free(ref_dbs);
  for (si = 0; si < seed_count; ++si) {
    /* @filter_efficiency_statistics */
    #ifdef ENABLE_STATS
      stats__display(ref_index[si], FILTER_STATS[si]);
    #endif
    seed__destroy(seeds[si]);
  }
#ifdef ENABLE_STATS
  stats__destroy();
#endif
  free(ref_index);
  free(seeds);
  ALIGNMENT__DESTROY(alignment);
  REV_SEQ__DESTROY(tmp_sequence);
  QUALITY_MASK__DESTROY
  VERB_FILTER(VERBOSITY_MODERATE, INFO__(("\nAll done in %ld seconds.\n", time(NULL) - start_time));)
  return RETURN_SUCCESS;
}
