#include "stats.h"

/**
 * Displays the traceback strings of the mapped reads (for statistics)
 */
void  display_tracebacks(GenomeMapType* genome_map) {
	static int traceback_patterns_index = 1;
	static char filename[256];
	sprintf(filename, "traceback_patterns%05d.txt", traceback_patterns_index);
	FILE* f = fopen(filename, "w");
	int i, ri, k, p;
	for (i = 0; i < genome_map->ref_db->size; ++i) {
		if (genome_map->g_map[i] && (ri = genome_map->g_map[i]->read_idx) != UNMAPPED) {
			for (k = 0; k < MAP_DETAIL_SIZE; ++k) {
				if (genome_map->hitmap->map[ri][k].ref_position == i) {
				  // print traceback
				  TRACEBACK_DISPLAY((genome_map->hitmap->map[ri][k].traceback), genome_map->hitmap->map[ri][k].traceback_seq_len, p, f);
				  fprintf(f, "\t%d", genome_map->hitmap->map[ri][k].score);
				  if ( genome_map->hitmap->map[ri][k].sense & (ALIGNMENT_SENSE_REVERSE) ) {
					fprintf(f, "\tR");
				  }
				  fprintf(f, "\n");
			      break;
				}
			}
		}
	}
	fflush(f);
	fclose(f);
	++traceback_patterns_index;
}

/**
 * Lists the unmapped reads,
 */
void list_unmapped_reads_translated(ReadsDBType* reads_db) {
	FILE* f = fopen("unmapped_reads_translated.txt", "w");
	int i, k;
	CODE_TYPE base, next_base;
	for (i = 0; i < reads_db->size; ++i) {
		if (reads_db->reads[i].disabled) {
			if (reads_db->reads[i].info) {
				fprintf(f, ">%s\n", reads_db->reads[i].info);
			} else {
				fprintf(f, ">%d\n", i);
			}
			for (base = 0; base < BASE_CODE_COUNT; ++base) {
				fprintf(f, "%c", BASE[base]);
				next_base = base;
				for (k = 0; k < reads_db->read_len; ++k) {
					next_base = base_compressed_transformation(next_base, reads_db->reads[i].sequence, k, 0);
					fprintf(f, "%c", BASE[next_base]);
				}
				fprintf(f, "\n");
			}
		}
	}
	fflush(f);
	fclose(f);
}

void list_unmapped_reads(ReadsDBType* reads_db) {
	FILE* f = fopen("unmapped_reads.txt", "w");
	int i;
	for (i = 0; i < reads_db->size; ++i) {
		if (reads_db->reads[i].disabled) {
			if (reads_db->reads[i].info) {
				fprintf(f, ">%s\n", reads_db->reads[i].info);
			} else {
				fprintf(f, ">%d\n", i);
			}
			fprintf(f, "%c", BASE[reads_db->reads[i].first_base]);
			display_compressed_code_sequence(reads_db->reads[i].sequence, reads_db->read_len, f);
		}
	}
	fflush(f);
	fclose(f);
}
