#ifndef UTIL_H_

#include "util.h"
#include <stdio.h>

int VERBOSITY = VERBOSITY_DEFAULT;

int ALIGNMENT_SENSE = DEFAULT_ALIGNMENT_SENSE;

inline void message(int verbosity_level, const char* msg) {
  VERB_FILTER(verbosity_level, printf("%s\n", msg););
}

/**
 * Displays a progress bar that shows the percentage of processed data
 */
inline void display_progress(long int p, long int total, long int selected) {
	static int p1 = -1, p2, MAX = 60;
	static char COMPLETED_SYMBOL = '=', UNCOMPLETED_SYMBOL = ' ', LAST_COMPLETED_SYMBOL = '>';
	if (p == 0 || (p = (p + 1)*100/total)  > p1) {
		p1 = p * MAX / 100;
		printf("\r\033[36;1m[");
		for (p2 = 0; p2 < p1; ++p2) {
			printf("%c", COMPLETED_SYMBOL);
		}
		if (p1 && p1 != MAX) {
			printf("\b%c", LAST_COMPLETED_SYMBOL);
		}
		for (p2 = p1; p2 < MAX; ++p2) {
			printf("%c", UNCOMPLETED_SYMBOL);
		}
		printf("]");
		printf(" %3d%%",  (int)min(p, 100));
		if (selected > 0) {
			printf("  >> %12ld - %5.2lf%%", selected, (selected *100.0/total));
		}
		printf("\033[0m");
		fflush(stdout);
		if (p1 == MAX) {
			printf("\n");
		}
		p1 = p;
	}
}

/**
 * Sort the first array (values) without changing it and put the indices in the right order
 * in the second array (result_indices).
 * @param values The values that need to be sorted
 * @param result_indices The array holding the indices of the sorted values
 * @param size The size of the two arrays
 */
inline void sort__indices(const int* values, int* result_indices, const int size) {
	int i, j, val_idx;
	for (i = 0; i < size; ++i) {
		result_indices[i] = i;
	}
    for (i = 1; i < size; ++i) {
	    val_idx = result_indices[i];
	    j = i - 1;
	    do {
	    	if (values[result_indices[j]] < values[val_idx]) {
	    		result_indices[j + 1] = result_indices[j];
	    		--j;
	    	} else {
	    		break;
	    	}
	    } while (j >= 0);
	    result_indices[j + 1] = val_idx;
    }
}

#endif /* UTIL_H_ */
