#ifndef UTIL_H_
#define UTIL_H_

#include <stdio.h>

#define PROGRAM_NAME "SToRM"

#define __w {printf("\033[35;1m%s %d\033[0m\n", __FILE__, __LINE__);fflush(stdout);}

#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))
#define abs(a)    ((a) >= 0  ? (a) :-(a))

#define VERBOSITY_NONE     0
#define VERBOSITY_MODERATE 1
#define VERBOSITY_HIGH     2
#define VERBOSITY_ANNOYING 3
#define VERBOSITY_MAX      VERBOSITY_ANNOYING
#define VERBOSITY_DEFAULT  VERBOSITY_MODERATE

#define DEFAULT_MAX_THREADS 8

int VERBOSITY;

int MAX_THREADS;

int ALIGNMENT_SENSE;

#define ALIGNMENT_SENSE_FORWARD 0x1
#define ALIGNMENT_SENSE_REVERSE 0x2

#define DEFAULT_ALIGNMENT_SENSE (ALIGNMENT_SENSE_FORWARD | ALIGNMENT_SENSE_REVERSE)


#define MD_COUNT(indels) ((indels < 2) ? 4 : ((indels < 4) ? 2 : 1))

/**
 * Displays a progress bar that shows the percentage of processed data
 */
inline void display_progress(long int p, long int total, long int selected);

#define VERB_FILTER(verbosity_level, f) if ((verbosity_level) <= VERBOSITY) { f; }

inline void message(int verbosity_level, const char* msg);


#define ERROR__(message)   {printf("\033[31;1m"); printf message; printf("\033[0m\n"); }
#define WARNING__(message) {printf("\033[33;1m"); printf message; printf("\033[0m\n"); }
#define INFO__(message)    {printf("\033[32;1m"); printf message; printf("\033[0m\n"); }
#define DEBUG__(message)   {printf("\033[35;1m"); printf message; printf("\033[0m\n"); }

#define SAFE_FAILURE__ALLOC(ptr, size, unit_type)\
  ptr = (unit_type*)calloc((size_t)size, (size_t)sizeof(unit_type));	\
	if (ptr == NULL) { \
	  ERROR__(("\nNot enough available memory.\n%s:%d  %lu X %lu = %lu bytes needed\n\nExiting.\n", __FILE__, __LINE__, (unsigned long int)size, (unsigned long int)sizeof(unit_type), (unsigned long int)(size) * sizeof(unit_type))) \
		exit(1); \
	}

/**
 * Sort the first array (values) without changing it and put the indices in the right order
 * in the second array (result_indices).
 * @param values The values that need to be sorted
 * @param result_indices The array holding the indices of the sorted values
 * @param size The size of the two arrays
 */
inline void sort__indices(const int* values, int* result_indices, const int size) ;

#endif /* UTIL_H_ */
